import tensorflow as tf
import math
from tensorx.utils import as_tensor


def identity(x, name: str = None) -> tf.Tensor:
    """ Identity function

    Returns a tensor with the same content as the input tensor.

    Args:
        x (`Tensor`): The input tensor.
        name (`str`): name for this op

    Returns:
        tensor (`Tensor`): of the same shape, type and content of the input tensor.
    """
    return tf.identity(x, name=name)


def sigmoid(x):
    """ Sigmoid function

    Element-wise sigmoid function, defined as:

    $$
    f(x) = \\frac{1}{1 + \\exp(-x)}
    $$

    Args:
        x (`Tensor`): A tensor or variable.

    Returns:
        A tensor (`Tensor`): with the result of applying the sigmoid function to the input tensor.
    """
    return tf.nn.sigmoid(x)


def hard_sigmoid(x, name="hard_sigmoid"):
    """ Hard Sigmoid

    Segment-wise linear approximation of sigmoid. (Faster than sigmoid)

    !!! note
        Approximates the sigmoid function in 3 parts: 0, scaled linear, 1.

        returns `0.` if `x < -2.5`, `1.` if `x > 2.5`.
        In `-2.5 <= x <= 2.5`, returns `0.2 * x + 0.5`.

    Args:
        x (`Tensor`): input tensor
        name (`str`): name for this op

    Returns:
        tensor (`Tensor): the result of applying an approximated element-wise sigmoid to the input tensor

    """
    x = as_tensor(x)
    with tf.name_scope(name):
        slope = as_tensor(0.2, x.dtype)
        shift = as_tensor(0.5, x.dtype)
        x = (slope * x) + shift
        zero = as_tensor(0., x.dtype)
        one = as_tensor(1., x.dtype)
        x = tf.clip_by_value(x, zero, one)

    return x


def tanh(x):
    """ Hyperbolic tangent (tanh) function.

    The element-wise hyperbolic tangent function is essentially a rescaled
    sigmoid function. The sigmoid function with range $[0,1]$ is defined as follows:

    $$
    f(x) = \\frac{1}{1 + \\exp(-x)}
    $$

    the hyperbolic tangent is a re-scaled function such that it's outputs range $[-1,1]$ defined as:
    $$
    tanh(x) = 2f(2x)−1
    $$

    which leads us to the standard definition of hyperbolic tangent

    $$
    tanh(x)=\\frac{e^{x}-e^{-x}}{e^{x}+e^{-x}}
    $$

    Args:
        x (`Tensor`): an input tensor

    Returns:
        tensor (`Tensor`): a tensor with the result of applying the element-wise hyperbolic tangent to the input

    """
    return tf.nn.tanh(x)


def relu(x):
    """ relu activation

    A Rectifier linear unit [1] is defined as:

    $$
    f(x)= \\max(0, x)
    $$

    !!! cite "References"
        1. (Vinod & Hinton, 2010) [Rectified linear units improve restricted boltzmann machines](https://www.cs.toronto.edu/~fritz/absps/reluICML.pdf)

    Args:
        x (`Tensor`):  input tensor

    Returns:
        tensor (`Tensor`) that results in element-wise rectifier applied to x.
    """
    return tf.nn.relu(x)


def elu(x, alpha=1.0):
    """ elu activation

    An Exponential Linear Unit (ELU) is defined as:

    $$
    f(x)=\\left\\{\\begin{array}{cc}x & x>0 \\\\
    \\alpha \\cdot \\left(e^{x}-1\\right) & x<=0
    \\end{array}\\right\\}
    $$

    !!! cite "References"
        1. (Clevert et al. 2015) [Fast and accurate deep network learning by exponential linear units (ELUs)](https://arxiv.org/abs/1511.07289).

    Args:
        x (`Tensor`): an input tensor
        alpha (`float`): A scalar, slope of positive section.

    Returns:
        tensor (`Tensor`): resulting from the application of the elu activation to the input tensor.
    """
    y = tf.nn.elu(x)
    if alpha == 1:
        return y
    else:
        return tf.where(x > 0, y, x * y)


def gelu(x, approximate: bool = True) -> tf.Tensor:
    """ Gaussian Error Linear Unit.

    Computes gaussian error linear:
        `0.5 * x * (1 + tanh(sqrt(2 / pi) * (x + 0.044715 * x^3)))` or
        `x * P(X <= x) = 0.5 * x * (1 + erf(x / sqrt(2)))`, where P(X) ~ N(0, 1),
        depending on whether approximation is enabled.

    !!! cite "References"
        1. [Gaussian Error Linear Units (GELUs)](https://arxiv.org/abs/1606.08415)
        2. [BERT](https://arxiv.org/abs/1810.04805).
    Args:
        x (`Tensor`):  Must be one of the following types:
            `float16`, `float32`, `float64`.
        approximate (bool): whether to enable approximation.

    Returns:
        tensor (`Tensor`): with the same type as `x`
    """
    x = tf.convert_to_tensor(x)
    if approximate:
        pi = tf.cast(tf.constant(math.pi), x.dtype)
        coefficient = tf.cast(0.044715, x.dtype)
        return 0.5 * x * (1.0 + tf.tanh(tf.sqrt(2.0 / pi) * (x + coefficient * tf.pow(x, 3))))
    else:
        return 0.5 * x * (1.0 + tf.math.erf(x / tf.cast(tf.sqrt(2.0), x.dtype)))


def selu(x):
    """ The Scaled Exponential Linear Unit (SELU)

    `scale * x` if `x > 0` and `scale * alpha * (exp(x) - 1)` if `x < 0`
    where alpha and scale are pre-defined constants (`alpha = 1.67326324` and `scale = 1.05070098`).

    The values of alpha and scale are chosen so that the mean and variance of the inputs are preserved
    between two consecutive layers as long as the weights are initialized correctly
    (see `variance_scaling_init`).

    To be used together with initializer = tf.variance_scaling_initializer(factor=1.0, mode='FAN_IN').
    For correct dropout, use tf.contrib.nn.alpha_dropout.

    !!! cite "References"
        1. [Self-Normalizing Neural Networks](https://arxiv.org/pdf/1706.02515.pdf)

    Args:
        x (`Tensor`): input tensor

    Returns:
        tensor (`Tensor`): results in `scale * x` if `x > 0` and `scale * alpha * (exp(x) - 1)` if `x < 0`
    """
    return tf.nn.selu(x)


def softmax(x, axis=None, name=None):
    """ softmax activation

    Softmax activation function, is equivalent to `softmax = tf.exp(logits) / tf.reduce_sum(tf.exp(logits), axis)`
    and it is defined as:

    $$
    \\sigma(\\mathbf{z})_{i}=\\frac{e^{z_{i}}}{\\sum_{j=1}^{K} e^{z_{j}}}
    $$

    Args:
        x (`Tensor`): input tensor
        axis (`int`): the dimension softmax would be performed on. The default is -1 which indicates the last dimension.
        name (`str`): name for this op

    Returns:
        tensor (`Tensor`): output resulting from the application of the softmax function to the input tensor

    """
    return tf.nn.softmax(x, axis=axis, name=name)


def sparsemax(logits, name: str = None) -> tf.Tensor:
    """Computes the sparsemax activation function [1]

    For each batch `i` and class `j` we have
      sparsemax[i, j] = max(logits[i, j] - tau(logits[i, :]), 0)

    References:
        [1]: https://arxiv.org/abs/1602.02068

    Args:
        logits (`Tensor`): tensor with dtype: `half`, `float32`,`float64`.
        name (`str`): A name for the operation (optional).

    Returns:
        tensor (`Tensor`): with the same type as the input logits.
    """

    with tf.name_scope(name, "sparsemax"):
        logits = tf.convert_to_tensor(logits, name="logits")
        obs = tf.shape(logits)[0]
        dims = tf.shape(logits)[1]

        z = logits - tf.reduce_mean(logits, axis=1)[:, tf.newaxis]

        # sort z
        z_sorted, _ = tf.nn.top_k(z, k=dims)

        # calculate k(z)
        z_cumsum = tf.cumsum(z_sorted, axis=1)
        k = tf.range(
            1, tf.cast(dims, logits.dtype) + 1, dtype=logits.dtype)
        z_check = 1 + k * z_sorted > z_cumsum
        # because the z_check vector is always [1,1,...1,0,0,...0] finding the
        # (index + 1) of the last `1` is the same as just summing the number of 1.
        k_z = tf.reduce_sum(tf.cast(z_check, tf.int32), axis=1)

        # calculate tau(z)
        indices = tf.stack([tf.range(0, obs), k_z - 1], axis=1)
        tau_sum = tf.gather_nd(z_cumsum, indices)
        tau_z = (tau_sum - 1) / tf.cast(k_z, logits.dtype)

        # calculate p
        return tf.maximum(
            tf.cast(0, logits.dtype), z - tau_z[:, tf.newaxis])


__all__ = [
    "identity",
    "sigmoid",
    "hard_sigmoid",
    "tanh",
    "relu",
    "selu",
    "elu",
    "gelu",
    "softmax",
    "sparsemax"
]



import tensorflow as tf
from typing import Union, List
from tensorx.utils import as_tensor
from tensorx.math import batch_sparse_dot, sparse_l2_norm
from tensorx.ops import grid_2d


def cosine_distance(tensor1, tensor2, dtype=tf.float32):
    """ cosine_distance

    Computes the pairwise cosine distance between two non-zero tensors on their last dimension.
    The cosine distance is defined as 1 - cosine similarity. With the cosine similarity defined as:

    $$
    similarity =\\cos (\\theta)=\\frac{\\mathbf{A} \\cdot \\mathbf{B}}{\\|\\mathbf{A}\\|\\|\\mathbf{B}\\|}=\\frac{
    \\sum_{i=1}^{n} A_{i} B_{i}}{\\sqrt{\\sum_{i=1}^{n} A_{i}^{2}} \\sqrt{\\sum_{i=1}^{n} B_{i}^{2}}}
    $$

    Args:
        tensor1 (`Tensor`): first tensor
        tensor2 (`Tensor`): second tensor
        dtype (`DType`): assumed type of both tensors

    Returns:
        distance (`Tensor`): the pairwise cosine distance between two tensors
    """
    tensor1 = tf.convert_to_tensor(tensor1, dtype)
    tensor2 = tf.convert_to_tensor(tensor2, dtype)

    dot_prod = tf.reduce_sum(tf.multiply(tensor1, tensor2), -1)
    norm1 = tf.norm(tensor1, axis=-1)
    norm2 = tf.norm(tensor2, axis=-1)

    norm12 = norm1 * norm2
    cos12 = dot_prod / norm12

    sim = tf.where(tf.math.is_nan(cos12), tf.zeros_like(cos12), cos12)

    # if we need to correct this to angular distance, acos(1.000001) is nan)
    sim = tf.clip_by_value(sim, -1., 1.)
    return 1 - sim


def euclidean_distance(tensor1, tensor2):
    """ Computes the euclidean distance between two tensors.

    The euclidean distance or $L^2$ distance between points $p$ and $q$ is the length of the line segment
    connecting them.

    $$
    distance(q,p) =\\sqrt{\\sum_{i=1}^{n}\\left(q_{i}-p_{i}\\right)^{2}}
    $$

    Args:
        tensor1: a ``Tensor``
        tensor2: a ``Tensor``
        dim: dimension along which the euclidean distance is computed

    Returns:
        ``Tensor``: a ``Tensor`` with the euclidean distances between the two tensors

    """
    tensor1 = tf.convert_to_tensor(tensor1)
    tensor2 = tf.convert_to_tensor(tensor2)

    distance = tf.sqrt(tf.reduce_sum(tf.square(tensor1 - tensor2), axis=-1))

    return distance


def sparse_euclidean_distance(sp_tensor, tensor2):
    """ Computes the euclidean distance between two tensors.

        Args:
            sp_tensor (`Union[Tensor,SparseTensor]`): a tensor or sparse tensor
            tensor2 (`Tensor`): a dense tensor

        Returns:
            distance (`Tensor`): euclidean distances between the two tensors

        """
    tensor1 = tf.SparseTensor.from_value(sp_tensor)
    if tensor1.values.dtype != tf.float32:
        tensor1.values = tf.cast(tensor1.values, tf.float32)
    tensor2 = tf.convert_to_tensor(tensor2)

    distance = tf.sqrt(tf.reduce_sum(tf.square(tensor1 - tensor2), axis=-1))

    return distance


def pairwise_euclidean_distance(tensor1, tensor2, keepdims=False):
    """ Computes the euclidean distance between two tensors.

    Args:
        tensor1 (`Tensor`): a dense tensor
        tensor2 (`Tensor`): a dense tensor
        keepdims (`Bool`): if True, the result maintains the dimensions of the original result

    Returns:
        distance (`Tensor`): euclidean distances between the two tensors
    """
    tensor1 = tf.convert_to_tensor(tensor1)
    tensor2 = tf.convert_to_tensor(tensor2)
    tensor1 = tf.expand_dims(tensor1, 1)

    distance = tf.sqrt(tf.reduce_sum(tf.square(tensor1 - tensor2), axis=-1, keepdims=keepdims))

    return distance


def torus_l1_distance(point, shape):
    """ Computes the l1 distance between a given point or batch of points and all other points in a torus

    Args:
        point (`Tensor`): a rank 0 or rank 1 tensor with the coordinates for a point or a rank 2 tensor with a batch of points.
        shape (`List`): a list with the shape for the torus - either 1D or 2D

    Returns:
        distances (`Tensor`): a rank 1 or 2 tensor with the distances between each point in the 1D torus and each unique
        coordinate in the shape

    Examples:
        * distance for a single point
        `torus_l1_distance(1,[4])`
        or
        `torus_1d_l1_distance([1],[4])`

        ```python
        [ 1.,  0.,  1.,  2.]
        ```

        * distance for multiple points `torus_l1_distance([[2],[3]],[4])`

        ```python
        [[ 2.,  1.,  0.,  1.],
         [ 1.,  2.,  1.,  0.]]
        ```

        * distance between a point and other coordinates in a 2D torus

        ```python
        r = torus_l1_distance([[1,1],[1,2]],[3,3])
        np.reshape(r,[-1,3,3])

        [[[ 2.,  1.,  2.],
          [ 1.,  0.,  1.],
          [ 2.,  1.,  2.]],

         [[ 2.,  2.,  1.],
          [ 1.,  1.,  0.],
          [ 2.,  2.,  1.]]]
        ```

    """
    point = as_tensor(point, tf.float32)
    if len(shape) == 1:
        max_x = shape[0]
        coor_x = tf.range(0, max_x, 1, dtype=tf.float32)
        dx = tf.abs(point - coor_x)
        distance = tf.minimum(dx, tf.math.mod(-dx, max_x))
    elif len(shape) == 2:
        max_x = shape[0]
        max_y = shape[1]

        xys = grid_2d(shape)
        xys = tf.cast(xys, tf.float32)

        xs, ys = tf.unstack(xys, num=2, axis=-1)

        px, py = tf.unstack(point, num=2, axis=-1)
        px = tf.expand_dims(px, 1)
        py = tf.expand_dims(py, 1)

        dx = tf.abs(px - xs)
        dy = tf.abs(py - ys)

        dx = tf.minimum(dx, tf.math.mod(-dx, max_x))

        dy = tf.minimum(dy, tf.math.mod(-dy, max_y))

        distance = dx + dy
    else:
        raise ValueError("Invalid shape parameter, shape must have len 1 or 2")

    return distance


def batch_manhattan_distance(tensor1, tensor2, keepdims=False):
    """ Compute the pairwise manhattan distance between a batch of tensors and a second tensor

    If any tensor is a ``SparseTensor``, it is converted to

    Args:
        tensor1 (`Union[Tensor,SparseTensor]`): a batch of tensors or sparse tensor
        tensor2 (`Union[Tensor,SparseTensor]`): another tensor or a sparse tensor
        keepdims (`Bool`): if True keeps the dimensions of the original tensors

    Returns:
        distance (`Tensor`): the manhattan distance between the two tensors

    """
    tensor1 = as_tensor(tensor1)
    tensor2 = as_tensor(tensor2)

    if isinstance(tensor1, tf.SparseTensor):
        tensor1 = tf.sparse.to_dense(tensor1)
    if isinstance(tensor2, tf.SparseTensor):
        tensor2 = tf.sparse.to_dense(tensor2)

    tensor1 = tf.expand_dims(tensor1, 1)
    abs_diff = tf.abs(tf.subtract(tensor1, tensor2))
    return tf.reduce_sum(abs_diff, axis=-1, keepdims=keepdims)


def batch_sparse_cosine_distance(sp_tensor, tensor, dtype=tf.float32, keepdims=False):
    """ Computes the cosine distance between two non-zero `SparseTensor` and `Tensor`

        Warning:
            1 - cosine similarity is not a proper distance metric, to repair the triangle inequality property while
            maintaining the same ordering, it is necessary to convert to angular distance

        Args:
            sp_tensor: a `SparseTensor`
            tensor: a `Tensor`
            dtype:
            keepdims: keeps the original dimension of the input tensor

        Returns:
            a `Tensor` with the cosine distance between two tensors
        """
    sp_tensor = as_tensor(sp_tensor, dtype)
    tensor = tf.convert_to_tensor(tensor, dtype)

    dot_prod = batch_sparse_dot(sp_tensor, tensor, keepdims=keepdims)

    norm1 = sparse_l2_norm(sp_tensor, axis=-1, keepdims=True)
    norm2 = tf.norm(tensor, axis=-1)

    norm12 = norm1 * norm2
    if keepdims:
        norm12 = tf.expand_dims(norm12, -1)

    cos12 = dot_prod / norm12

    sim = tf.where(tf.math.is_nan(cos12), tf.zeros_like(cos12), cos12)
    sim = tf.clip_by_value(sim, -1., 1.)

    return 1 - sim


def sinkhorn(tensor1, tensor2, epsilon, n_iter, cost_fn=None):
    """ Sinkhorn Distance

    !!! info
        Optimal Transport (OT) provides a framework from which one can define a more powerful geometry to compare
        probability distributions. This power comes, however, with a heavy computational price. The cost of computing OT
        distances scales at least in $O(d^3 log(d))$ when comparing two histograms of dimension $d$. Sinkhorn algorithm
        alleviate this problem by solving an regularized OT in linear time.

    Given two measures with n points each with locations x and y
    outputs an approximation of the Optimal Transport (OT) cost with regularization
    parameter epsilon, niter is the maximum number of steps in sinkhorn loop

    !!! cite "References" 1. [Concerning nonnegative matrices and doubly stochastic matrices](
    https://msp.org/pjm/1967/21-2/p14.xhtml) 2. [Sinkhorn Distances:Lightspeed Computation of Optimal Transport](
    https://papers.nips.cc/paper/4927-sinkhorn-distances-lightspeed-computation-of-optimal-transport.pdf)

    Args:
        tensor1 (`Tensor`): a tensor representing a distribution
        tensor2 (`Tensor`): other tensor with another distribution
        epsilon (float): regularization term >0
        n_iter (`int`): number of sinkhorn iterations
        cost_fn (`Callable`): function that returns the cost matrix between y_pred and y_true, defaults to $|x_i-y_j|^p$.

    Returns:
        cost (`Tensor`): sinkhorn cost of moving from the mass from the model distribution `y_pred` to the empirical
        distribution `y_true`.
    """

    def cost_matrix(x, y, p=2):
        """ cost matrix of $|x_i-y_j|^p$.
        """
        xc = tf.expand_dims(x, 1)
        yr = tf.expand_dims(y, 0)
        d = tf.math.pow(tf.abs(xc - yr), p)
        return tf.reduce_sum(d, axis=-1)

    # n x n Wasserstein cost function
    if cost_fn is None:
        cost_m = cost_matrix(tensor1, tensor2)
    else:
        cost_m = cost_fn(tensor1, tensor2)
    # both marginals are fixed with equal weights
    # mu = Variable(1. / n * tf.ones([n], dtype=tf.float32))
    # nu = Variable(1. / n * tf.ones([n], dtype=tf.float32))

    n = tf.shape(tensor1)[0]

    init_v = tf.cast(n, tf.float32) * tf.ones([n], dtype=tf.float32)

    mu = 1. / init_v
    nu = 1. / init_v

    # Parameters of the sinkhorn algorithm.
    rho = 1  # (.5) **2 # unbalanced transport
    tau = -.8  # nesterov-like acceleration
    lam = rho / (rho + epsilon)  # Update exponent
    thresh = 0.1  # stopping criterion

    # Elementary operations .....................................................................
    def ave(u, u1):
        # Barycenter subroutine, used by kinetic acceleration through extrapolation.
        return tau * u + (1 - tau) * u1

    def M(u, v):
        # Modified cost for logarithmic updates $M_{ij} = (-c_{ij} + u_i + v_j) / \epsilon$
        return (-cost_m + tf.expand_dims(u, 1) + tf.expand_dims(v, 0)) / epsilon

    def lse(A):
        # log-sum-exp
        return tf.reduce_logsumexp(A, axis=1, keepdims=True)

    # Actual Sinkhorn loop ......................................................................
    init_u, init_v, init_err = 0. * mu, 0. * nu, 0.
    actual_nits = 0  # to check if algorithm terminates because of threshold or max iterations reached

    def body(i, u, v, _):
        u0 = u  # to check the error threshold
        new_u = epsilon * (tf.math.log(mu) - tf.squeeze(lse(M(u, v)))) + u
        new_v = epsilon * (tf.math.log(nu) - tf.squeeze(lse(tf.transpose(M(new_u, v))))) + v
        # accelerated unbalanced iterations
        # u = ave( u, lam * ( epsilon * ( torch.log(mu) - lse(M(u,v)).squeeze()   ) + u ) )
        # v = ave( v, lam * ( epsilon * ( torch.log(nu) - lse(M(u,v).t()).squeeze() ) + v ) )
        error = tf.reduce_sum(tf.abs(new_u - u0))

        return i + 1, new_u, new_v, error

    def cond(i, u, v, err):
        return tf.logical_and(tf.less(err, thresh), tf.less(i, n_iter))

    i, u, v, err = tf.while_loop(cond=cond,
                                 body=body,
                                 loop_vars=(0, init_u, init_v, init_err))

    pi = tf.exp(M(u, v))  # Transport plan p_i = diag(a)*K*diag(b)

    return tf.reduce_sum(pi * cost_m)  # pi * cost_m  # tf.reduce_sum(pi * cost_m)


__all__ = [
    "torus_l1_distance",
    "sparse_euclidean_distance",
    "euclidean_distance",
    "cosine_distance",
    "batch_manhattan_distance",
    "batch_sparse_cosine_distance",
    "sinkhorn",
    "pairwise_euclidean_distance"
]



import tensorflow as tf
import numpy as np
import logging
from collections import Counter

logging.captureWarnings(True)  # captures into py.warnings
logger = logging.getLogger('tensorx')
logger.setLevel(logging.DEBUG)
import re


class Graph:
    """ Graph

    Simple append-only graph data structure. It keeps track of nodes, directed edges, and endpoint nodes.

    Note:
        A node without edges counts as both an input and output node of the graph

    Attributes:
        nodes(set): set of node objects (object need tobe hashable)
        edges_in(dict): a dictionary that maps nodes in the keys to a list of notes with edges coming into that node
        edges_out(dict): a dictionary that maps nodes in the keys to a list of notes with edges coming from that node

        in_nodes(dict): key-only dictionary (ordered set) with input nodes (nodes without input edges).
        out_nodes(dict): key-only dictionary (ordered set) with output nodes of the graph (nodes without output edges)
    """

    def __init__(self):
        self.nodes = set()
        self.in_nodes = dict()
        self.out_nodes = dict()
        self.edges_in = dict()
        self.edges_out = dict()

    def add_node(self, node):
        if node not in self.nodes:
            self.nodes.add(node)
            self.edges_in[node] = []
            self.edges_out[node] = []

            # dict without values is a special case of ordered set
            self.in_nodes[node] = None
            self.out_nodes[node] = None

    def add_edge(self, node1, node2):
        """ Adds a new edge to the graph and removes
         the nodes from input roots or output leafs
         if these are changed


        Args:
            node1:
            node2:

        Returns:

        """
        self.add_node(node1)
        self.add_node(node2)
        self.edges_out[node1].append(node2)
        self.edges_in[node2].append(node1)

        # update endpoints
        if node1 in self.out_nodes:
            del self.out_nodes[node1]
        if node2 in self.in_nodes:
            del self.in_nodes[node2]

    def dependency_iter(self):
        """ returns a dictionary with a map from nodes to dependency priorities
            with lower values having higher priority. Keys are ordered by priority from
            lower to higher and number of dependencies from lower to higher

            Notes:
               Transversing a graph by priority guarantees that when we visit a node
               all it's dependencies have already been visited, additionally, ordering by
               number of dependencies guarantees that we can maintain a minimum result
               cache when transversing the graph.

           Returns:
               nodes (`dict`): dictionary from nodes to (priorities,number of dependencies)
        """
        priority = dict()
        visited = set()
        nodes = list(self.in_nodes)

        while nodes:
            current = nodes.pop(0)
            visited.add(current)
            delayed = False

            if not self.edges_in[current]:
                priority[current] = (0, len(self.edges_out[current]))
            else:
                # delay node if not all dependencies are ready
                if any([dep not in priority for dep in self.edges_in[current]]):
                    nodes.append(current)
                    delayed = True
                else:
                    priorities = [priority[dep][0] for dep in self.edges_in[current]]
                    priority[current] = (max(priorities) + 1, len(self.edges_out[current]))

            if not delayed:
                next_nodes = self.edges_out[current]
                for next_node in next_nodes:
                    if next_node not in visited:
                        nodes.insert(0, next_node)

        sorted_priority = dict(sorted(priority.items(), key=lambda kv: kv[1], reverse=False))

        return sorted_priority

    # TODO this doesn't take Tensors, only layers
    @staticmethod
    def build(inputs, outputs, add_missing_inputs=False):
        """ build_graph

        #TODO perhaps I could change the name to inputs, to make this more general, any node with inputs would work
        Transverses nodes starting from outputs and following `node.input_layers` to create the edges.

        !!! note
            use `add_missing_inputs` if you have graph inputs but might have other dependencies that might not have
            been created explicitly. Example: in an RNN layer, if a previous state is not passed explicitly, a default
            one is created by the layer and stored in input layers. You might be aware of this input node to a graph but
            not want to pass it explicitly to inputs.

        Args:
            inputs: input terminal layers where the graph must stop
            outputs: output layers from which we start to populate the graph
            add_missing_inputs: if True and `inputs` are provided, input nodes found that are not in given inputs
                will be added to the graph. If False ValueError is raised with a list of inputs not specified
                (missing dependencies).

        Returns:
            graph (`Graph`): a graph from the outputs to the given input, or to every input found if these are not
                specified.
        """
        graph = Graph()
        inputs = dict.fromkeys(as_list(inputs))
        graph.outputs = as_list(outputs)
        outputs = dict.fromkeys(as_list(outputs))

        # add terminals to the graph
        # for layer in input_layers:
        #    graph.add_node(layer)
        for layer in outputs:
            graph.add_node(layer)

        dependencies = dict()
        missing_dependencies = dict()

        def add_dep(out, dep, target: dict):
            if out not in target:
                target[out] = set()
            target[out].add(dep)

        # arg order in a path to the output
        arg_ord = {out: (0,) for out in outputs}
        visited = set()
        node_queue = list(zip(outputs, outputs))

        while node_queue:
            current_node, target_output = node_queue.pop(0)
            if current_node not in visited:
                next_nodes = current_node.inputs
                if not next_nodes:
                    add_dep(target_output, current_node, dependencies)
                    if len(inputs) > 0 and current_node not in inputs and not add_missing_inputs:
                        add_dep(target_output, current_node, missing_dependencies)
                else:
                    if current_node in inputs:
                        add_dep(target_output, current_node, dependencies)
                    else:
                        for i, input_node in enumerate(next_nodes):
                            graph.add_edge(input_node, current_node)
                            node_queue.append((input_node, target_output))

                            # record arg order
                            if input_node not in arg_ord:
                                arg_ord[input_node] = arg_ord[current_node] + (i + 1,)

                visited.add(current_node)

        if any(missing_dependencies) and not add_missing_inputs:
            failed_str = []
            for output_layer in missing_dependencies:
                missing_str = "\n\t\t".join(map(str, missing_dependencies[output_layer]))
                failed_str.append(f"\t{str(output_layer)}: \n\t\t{missing_str}")
            failed_str = "\n".join(failed_str)

            raise ValueError(f"output layers missing inputs: \n {failed_str}")

        if inputs:
            missing_from_graph = list(filter(lambda x: x not in graph.in_nodes, inputs))

            if missing_from_graph:
                input_str = "\n\t ".join(map(str, missing_from_graph))
                output_str = "\n\t ".join(map(str, outputs))
                raise ValueError(f"no path between the output layers:\n\t {output_str} \n and input layers: \n\t"
                                 f"{input_str}")

        inputs.update(graph.in_nodes)
        outputs.update(graph.out_nodes)

        # if no ordered input is given
        # re-order by argument ordering
        if not inputs:
            # sort according to argument ordering from the outputs
            sorted_inputs = sorted(inputs, key=lambda in_layer: arg_ord[in_layer], reverse=False)
            inputs = dict.fromkeys(sorted_inputs)

        graph.in_nodes = inputs
        graph.out_nodes = outputs

        return graph

    def as_function(self, ord_inputs=None, ord_outputs=None, name="compiled_graph", compile=True):
        """ compiles the graph into a tensorflow callable compiled graph

        the idea is to use exec to create a function and then call tf.function
        on the created function. this is to avoid using loops and lists to
        run through the Graph instance and call compute.

        function parameters:
            converts all the non-constant Input Layer nodes into graph arguments which means that only Input
            layers are converted to params

        run through each layer and queue up nodes starting on the input
        write down a function as a series of compute calls with inputs from the previous
        layer outputs

        !!! note
            Another way to feed inputs to a graph is to use input layers

                ```python
                input_layer.value = in0
                input_Layer.value = in1
                outputs = graph()
                ```

            this adds a bit of a overhead since we have to write to the variable

        !!! bug "Dev Note"
            adding indices to the variables in the created function is necessary to avoid users
            naming layers like built-in identifiers.

        Args:
            ord_inputs: list of input that determines the order of resulting function arguments
            ord_outputs: list of outputs used to determine the return order
            name (`str`): function name, must be a valid python function name
            compile (`bool`): if True, returns a tensorflow graph else returns a python function

        Returns:
            function (`function`): an optimized TensorFlow static graph as a callable function or a python function

        """

        clean = lambda name_str: re.sub(r"\W|^(?=\d)", "_", name_str)
        name = clean(name)

        graph = self

        if not graph.out_nodes:
            raise ValueError("can't compile an empty graph")
        ord_inputs = as_list(ord_inputs)
        ord_outputs = as_list(ord_outputs)
        ord_nodes = list(self.dependency_iter())

        input_set: set = set(graph.in_nodes)
        if ord_inputs and not input_set.issuperset(ord_inputs):
            raise ValueError("all feedable_inputs must be part of the graph inputs")
        output_set: set = set(graph.out_nodes)
        if ord_outputs and len(output_set.difference(ord_outputs)) > 0:
            raise ValueError("all outputs must be part of the graph outputs")

        # if no input order is specified use the graph endpoint order
        outputs = dict.fromkeys(ord_outputs) if ord_outputs else graph.out_nodes

        # if we don't provide inputs it will just treat them as callables
        inputs = dict.fromkeys(ord_inputs) if ord_inputs else []  # graph.in_nodes

        # check if they are all dynamic inputs
        # in py3.7 the dict is an ordered set if we convert it back to a list
        node_index = list(range(len(graph.nodes)))

        feedable_inputs = list(inputs)
        node_map = {}
        for in_layer in feedable_inputs:
            layer_i = node_index.pop(0)
            in_name = in_layer.name.replace('/', '__')
            layer_name = f"{in_name}_{layer_i}"
            node_map[in_layer] = layer_name

        args_str = ", ".join(node_map.values())
        def_str = f"def {name}({args_str}):\n"
        other_str = []

        # all other inputs that are not feedable
        other_inputs = list(input_set.difference(feedable_inputs))
        node_map.update({in_layer: f"{in_layer.name}_{node_index.pop(0)}" for in_layer in other_inputs})

        # requires outer access to layers var
        for x in other_inputs:
            other_str.append(f"\t{node_map[x]} = layers[\"{node_map[x]}\"].compute()")

        other_str = "\n".join(other_str) + "\n" if other_str else ""

        # remove inputs
        # node_map contains input_nodes at this point
        for _ in range(len(node_map)):
            ord_nodes.pop(0)

        compute_str = []
        for current_node in ord_nodes:
            node_name = current_node.name.replace('/', '__')
            node_map[current_node] = f"{node_name}_{node_index.pop(0)}"
            node_name = node_map[current_node]
            # when layers have the same layer repeated as input, this causes problems
            # it's better to use the same input_layers as declared in the graph
            # dict from keys creates an ordered set which is not what we want
            # next_nodes = dict.fromkeys(graph.edges_in[current_node])
            next_nodes = graph.edges_in[current_node]
            in_args = ", ".join([node_map[node] for node in next_nodes])
            compute_str.append(f"\t{node_name} = layers[\"{node_name}\"].compute({in_args})")

        compute_str = "\n".join(compute_str)

        return_str = "\n\treturn {output_str}\n".format(output_str=", ".join([node_map[out] for out in outputs]))

        full_fn_str = def_str + other_str + compute_str + return_str
        logger.log(logging.DEBUG, f"converted function:\n {'-' * 10}\n\n {full_fn_str} \n{'-' * 10}")
        # layer map (for the closure above)
        # we feed the locals so that layers gets available in the above function
        layers = {v: k for k, v in node_map.items()}
        exec(full_fn_str, locals())
        fn = eval(name)
        fn.__doc__ = f"""{name}\n```python\n{full_fn_str}\n```"""

        if compile:
            fn = tf.function(fn)

        return fn

    def as_function_v2(self,
                       ord_inputs=None,
                       ord_outputs=None,
                       fn_name="compiled_graph",
                       stateful_inputs=False,
                       compile=True):
        # this could be static
        graph = self
        fn_name = re.sub(r"\W|^(?=\d)", "_", fn_name)

        if not graph.out_nodes:
            raise ValueError("can't compile an empty graph")
        ord_inputs = as_list(ord_inputs)
        ord_outputs = as_list(ord_outputs)
        ord_nodes = list(self.dependency_iter())

        input_set: set = set(graph.in_nodes)
        if ord_inputs and not input_set.issuperset(ord_inputs):
            raise ValueError("all feedable_inputs must be part of the graph inputs")
        output_set: set = set(graph.out_nodes)
        if ord_outputs and len(output_set.difference(ord_outputs)) > 0:
            raise ValueError("all outputs must be part of the graph outputs")

        # if no input order is specified use the graph endpoint order
        outputs = dict.fromkeys(ord_outputs) if ord_outputs else graph.out_nodes

        # if we don't provide inputs it will just treat them as callables
        inputs = dict.fromkeys(ord_inputs) if ord_inputs else []  # graph.in_nodes

        # check if they are all dynamic inputs
        # in py3.7 the dict is an ordered set if we convert it back to a list
        node_index = list(range(len(graph.nodes)))

        feedable_inputs = list(inputs)
        node_map = {}
        for in_layer in feedable_inputs:
            layer_i = node_index.pop(0)
            in_name = in_layer.name.replace('/', '__')
            layer_name = f"{in_name}_{layer_i}"
            node_map[in_layer] = layer_name

        # TODO check if these can be inputs that are not Input layers?
        # if using stateful inputs, we need to be able to pass no args
        fn_args = []
        update_inputs = []
        for layer, layer_name in node_map.items():
            if stateful_inputs and feedable_inputs:
                fn_args.append(f"{layer_name}=None")
                if not layer.constant:
                    update_inputs.append(f"\tif {layer_name}:\n\t\tlayers[\"{layer_name}\"].value = {layer_name}"
                                         f"\n\telse:\n"
                                         f"\t\t{layer_name} = layers[\"{layer_name}\"].compute()")
                else:
                    update_inputs.append(f"\tif not {layer_name}:\n"
                                         f"\t\t{layer_name} = layers[\"{layer_name}\"].compute()")

            else:
                fn_args.append(f"{layer_name}")

        args_str = ", ".join(fn_args)
        # args_str = ", ".join(node_map.values())
        def_str = f"def {fn_name}({args_str}):\n"

        update_str = "\n".join(update_inputs) + "\n"

        other_str = []

        # all other inputs that are not feedable
        other_inputs = list(input_set.difference(feedable_inputs))
        node_map.update({in_layer: f"{in_layer.name}_{node_index.pop(0)}" for in_layer in other_inputs})

        # requires outer access to layers var
        for x in other_inputs:
            other_str.append(f"\t{node_map[x]} = layers[\"{node_map[x]}\"].compute()")

        other_str = "\n".join(other_str) + "\n" if other_str else ""

        # remove inputs
        # node_map contains input_nodes at this point
        for _ in range(len(node_map)):
            ord_nodes.pop(0)

        compute_str = []
        for current_node in ord_nodes:
            name = current_node.name.replace('/', '__')
            node_map[current_node] = f"{name}_{node_index.pop(0)}"
            name = node_map[current_node]
            # when layers have the same layer repeated as input, this causes problems
            # it's better to use the same input_layers as declared in the graph
            # dict from keys creates an ordered set which is not what we want
            # next_nodes = dict.fromkeys(graph.edges_in[current_node])
            next_nodes = graph.edges_in[current_node]
            in_args = ", ".join([node_map[node] for node in next_nodes])
            compute_str.append(f"\t{name} = layers[\"{name}\"].compute({in_args})")

        compute_str = "\n".join(compute_str)

        return_str = "\n\treturn {output_str}\n".format(output_str=", ".join([node_map[out] for out in outputs]))

        full_fn_str = def_str + update_str + other_str + compute_str + return_str
        logger.log(logging.DEBUG, f"converted function:\n {'-' * 10}\n\n {full_fn_str} \n{'-' * 10}")

        # layer map (for the closure above)
        # we feed the locals so that layers gets available in the above function
        layers = {v: k for k, v in node_map.items()}
        print("\n", full_fn_str)
        exec(full_fn_str, locals())
        fn = eval(fn_name)
        fn.__doc__ = f"""{fn_name}\n```python\n{full_fn_str}\n```"""

        if compile:
            fn = tf.function(fn)

        return fn

    def draw(self, path):
        try:
            from pygraphviz import AGraph

            def vizstyle(layer_obj):
                # HTML for record nodes https://graphviz.org/doc/info/shapes.html#top

                dtype = layer_obj.dtype.name if layer_obj.dtype is not None else None

                label = f"<<TABLE BORDER=\"0\"" \
                        f"        CELLPADDING=\"2\"" \
                        f"        CELLSPACING=\"0\">" \
                        f"<TR><TD BGCOLOR=\"BLACK\"" \
                        f"        BORDER=\"1\"" \
                        f"        COLOR=\"BLACK\"" \
                        f"        VALIGN=\"BOTTOM\">" \
                        f"<FONT COLOR=\"WHITE\"><B>{type(layer_obj).__name__}</B></FONT>" \
                        f"</TD><TD BORDER=\"1\">{layer_obj.name}</TD></TR>" \
                        f"<TR>" \
                        f"<TD BORDER=\"1\"" \
                        f"    BGCOLOR=\"#aec4c7\"" \
                        f"    COLOR=\"BLACK\"" \
                        f"    ALIGN=\"RIGHT\">" \
                        f"units" \
                        f"</TD><TD BORDER=\"1\"" \
                        f"         COLOR=\"BLACK\"" \
                        f"         ALIGN=\"LEFT\">" \
                        f"{layer_obj.n_units}</TD></TR>" \
                        f"<TR>" \
                        f"<TD BORDER=\"1\" " \
                        f"    BGCOLOR=\"#aec4c7\"" \
                        f"    ALIGN=\"RIGHT\">dtype</TD>" \
                        f"<TD BORDER=\"1\"" \
                        f"    ALIGN=\"LEFT\">{dtype}</TD></TR>" \
                        f"</TABLE>>"
                return label

            viz_graph = AGraph(directed=True)

            for node in self.dependency_iter():
                if node not in viz_graph:
                    viz_graph.add_node(node.name, shape="none", margin=0,
                                       label=vizstyle(node))  # , label=vizstyle(node))
                for other_node in self.edges_out[node]:
                    viz_graph.add_edge(node.name, other_node.name)

            dependencies = dependency_graph(self.nodes)
            state_deps = [node for node in dependencies.nodes if not dependencies.edges_in[node] and
                          len(dependencies.edges_out[node]) > 1]
            for node in state_deps:
                if hasattr(node, "name"):
                    name = node.name
                else:
                    # might be a tf.Variable
                    try:
                        name = node.deref().name
                    except AttributeError as e:
                        name = str(type(node).__name__)

                viz_graph.add_node(name, color="red", shape="box")
                layers = dependencies.edges_out[node]
                for layer in layers:
                    viz_graph.add_edge(name, layer.name, color="red")

            viz_graph.layout(prog="dot")
            viz_graph.draw(path=path)

        except ImportError:
            raise ImportError("Could't find required pygraphviz module")

    def compute(self, *input_values):
        """ computes the graph output values based on the given input values

        Args:
            *input_values: input values with the same order as the graph inputs, or a dictionary mapping values to
            input layers.

        Returns:
            a tuple with the values for the correspondent graph outputs
        """
        if len(input_values) == 1 and isinstance(input_values[0], dict):
            input_dict = input_values[0]
            missing = list(filter(lambda x: x not in self.in_nodes, input_dict.keys()))

            if missing:
                missing_str = '\n\t'.join([f"{str(x)}" for x in missing])
                raise ValueError(f"inputs not found in graphs:\n"
                                 f"\t{missing_str}")

            ord_inputs = dict.fromkeys(list(self.in_nodes)[:len(input_dict)])
            input_values = [as_tensor(input_dict[input_layer]) for input_layer in ord_inputs]

        elif len(input_values) > len(self.in_nodes):
            raise ValueError(f"too many inputs:\n"
                             f"\tgraph expects {len(self.in_nodes)} inputs\n"
                             f"\tinputs passed {len(input_values)}")
        else:
            input_values = [as_tensor(input_value) for input_value in input_values]

        ord_inputs = dict.fromkeys(list(self.in_nodes)[:len(input_values)])
        input_dict = dict(zip(ord_inputs.keys(), input_values))
        other_inputs = set(self.in_nodes).difference(ord_inputs)

        node_iter = self.dependency_iter()

        result_cache = dict()
        visited = set()

        for node in node_iter:
            if node in input_dict:
                result_cache[node] = input_dict[node]
            elif node in other_inputs:
                result_cache[node] = node.compute()
            else:
                visited.add(node)

                # get input_node result, clean cache when no more dependencies on the same input
                def get_args(node):
                    args = []
                    ins = self.edges_in[node]
                    for in_node in ins:
                        res = result_cache[in_node]
                        priority, num_deps = node_iter[node]
                        node_iter[node] = (priority, num_deps - 1)
                        if num_deps - 1 == 0:
                            del result_cache[in_node]
                        args.append(res)
                    return args

                args = get_args(node)
                result_cache[node] = node.compute(*args)
                # result_cache[node] = node(*args)

        return tuple(map(lambda x: result_cache[x], self.out_nodes))

    def __call__(self, *input_values):
        return self.compute(*input_values)

    @classmethod
    def eval(cls, *layers):
        graph = Graph.build(inputs=None, outputs=layers)
        return graph()


def dependency_graph(layers):
    g = Graph()
    for layer in layers:
        state = layer.layer_state
        layer_vars = layer.variables

        # g.add_edge(state, layer)
        for var in layer_vars:
            ref = var.ref()
            g.add_edge(ref, layer)
    return g


def as_numerical_shape(shape: tf.TensorShape):
    return [-1 if dim is None else dim for dim in shape]


def as_tensor(x, dtype=None):
    """ Converts to tensor and casts to a given type if possible

    Args:
        x: an input ``Tensor``.
        dtype: the type we which to cast the input tensor into

    Returns:
        ``Tensor``: a tensor with the given dtype
    """
    if dtype is not None:
        dtype = tf.dtypes.as_dtype(dtype)
    if not isinstance(x, tf.SparseTensor):
        x = tf.convert_to_tensor(x)

    if dtype is not None:
        if x.dtype != dtype:
            x = tf.cast(x, dtype)
    return x


def as_list(items):
    """ Returns a list from one or multiple elements.

    if one element is passed, returns a list with one element,
    if a list,a tuple or a dictionary of elements is passed,
    returns a list with the elements or the keys if the input is a dict

    Note: we exclude SparseTensorValue because it is a named tuple
    and we want to feed the whole object as a single data sample if needed

    Args:
        items: one item, a tuple of elements or a list of elements

    Returns:
        a :obj:`list` with the elements in items
    """
    if items is None:
        items = []
    elif isinstance(items, (list, tuple, dict)) and not isinstance(items, (
            tf.compat.v1.SparseTensorValue, tf.SparseTensor)):
        items = list(items)
    else:
        items = [items]
    return items


def cast_like(x, y):
    """Cast x to y's dtype, if necessary."""
    x = tf.convert_to_tensor(x)
    y = tf.convert_to_tensor(y)

    if x.dtype.base_dtype == y.dtype.base_dtype:
        return x

    cast_x = tf.cast(x, y.dtype)
    if cast_x.device != x.device:
        x_name = "(eager Tensor)"
        try:
            x_name = x.name
        except AttributeError:
            pass
    return cast_x


def fix_reshape_dimensions(original_shape, target_shape):
    """ Find and replace a missing dimension in a target shape.

    Args:
      original_shape (`List[int]`): shape of tensor being reshaped
      target_shape (`List`): desired shape with at most a single -1
        which indicates a dimension that should be derived from the input shape.

    Returns:
        new_shape (`List`): the new shape with a -1 replaced with its computed value.

    Raises:
      `ValueError`: if the total tensor size of the new_shape is different than
      the original_shape, or more than one unknown (-1) dimension are specified.
    """
    target_shape = list(target_shape)

    target_n = 1
    target_unknown = None
    for i, dim in enumerate(target_shape):
        if dim < 0:
            if target_unknown is None:
                target_unknown = i
            else:
                raise ValueError('Can only specify one unknown dimension.')
        else:
            target_n *= dim

    msg = ('total size of new tensor must be unchanged, '
           'input_shape = {}, output_shape = {}'
           .format(original_shape, target_shape))

    # the target should match original known
    original_known = [dim if dim else 1 for dim in original_shape]
    original_n = np.prod(original_known, dtype=int, keepdims=False)
    if target_unknown is not None:
        if target_n == 0 or original_n % target_n != 0:
            raise ValueError(msg)
        elif None in original_shape and len(original_shape) != len(target_shape) and original_n > target_n:
            target_shape[target_unknown] = None
        else:
            target_shape[target_unknown] = original_n // target_n
    elif original_n != target_n:
        raise ValueError(msg)
    return target_shape



"""
    Initializers allow you to pre-specify an initialization strategy, encoded in
    the Initializer object, without knowing the shape and dtype of the variable
    being initialized.

    !!! Note
        Despite most initializers being accessible through Tensorflow, we decided to gather
        them here for the sake of namespace consistency (not all TF initializers are in tf.initializers)
"""
import tensorflow as tf
from typing import Callable, Optional


def zeros_init():
    """ Zeroes Initializer

    Initializer that generates tensors initialized to 0.

    Returns:
        initializer (Callable): an initializer that returns a tensor filled with 0 when called on a given shape.

    """
    return tf.zeros_initializer()


def ones_init():
    """ Ones Initializer

    Initializer that generates tensors initialized to 1.

    Returns:
        initializer (Callable): an initializer that returns a tensor filled with 1 when called on a given shape.

    """
    return tf.ones_initializer()


def constant_init(value=0):
    """ Constant Initializer

    The resulting tensor is populated with values of type dtype, as specified by arguments value
    following the desired shape.

    The argument value can be a constant value, or a list of values of type dtype. If value is a list, then the length
    of the list must be less than or equal to the number of elements implied by the desired shape of the tensor.
    In the case where the total number of elements in value is less than the number of elements required by the tensor
    shape, the last element in value will be used to fill the remaining entries. If the total number of elements in
    value is greater than the number of elements required by the tensor shape, the initializer will raise a ValueError.

    Args:
        value: A Python scalar, list or tuple of values, or a N-dimensional numpy array. All elements of
        the initialized variable will be set to the corresponding value in the value argument.

    Returns:
        initializer (Callable): an initializer that returns a tensor from the given specification and a given shape
    """
    return tf.constant_initializer(value)


def uniform_init(minval: float = -0.05, maxval: float = 0.05, seed=None):
    """ Random Uniform Initializer

    Initializer that generates tensors with a uniform distribution.

    Args:
        minval:  Lower bound of the range of random values to generate.
        maxval: Upper bound of the range of random values to generate. Defaults to 1 for float types.
        seed (int32/int64): seed for random number generator

    Returns:
         initializer (Callable): an initializer that returns a tensor from the given specification and a given shape

    """
    return tf.random_uniform_initializer(minval=minval, maxval=maxval, seed=seed)


def normal_init(mean: float = 0.0, stddev=0.05, seed=None):
    """ Random Normal Initializer

    Initializer that generates tensors with a normal distribution.

    Args:
        mean: Mean of the random values to generate.
        stddev: Standard deviation of the random values to generate.
        seed (int32/int64): seed for random number generator

    Returns:
         initializer (Callable): an initializer that returns a tensor from the given specification and a given shape

    """
    return tf.random_normal_initializer(mean=mean, stddev=stddev, seed=seed)


def glorot_uniform_init(seed: Optional = None) -> Callable:
    """ Glorot Uniform Initializer

    This initialisation keeps the scale of the gradients roughly the same in all layers to
    mitigate `vanishing` and `exploding gradients` see [1].
    
    References:
        [1] (Glorot and Bengio 2010), "Understanding the difficulty of training deep
        feedforward neural networks".
    
    Args:
        seed (int32/int64): seed for random number generator
    
    Returns:
        initializer (Callable): callable that creates an initial value from a given shape
    """
    return tf.initializers.glorot_uniform(seed)


def glorot_normal_init(seed: Optional = None) -> Callable:
    """ Glorot Normal Initializer

    This initialisation keeps the scale of the gradients roughly the same in all layers to
    mitigate `vanishing` and `exploding gradients` see [1].

    Draws samples from a truncated normal distribution.

    References:
        [1] (Glorot and Bengio 2010), "Understanding the difficulty of training deep
        feedforward neural networks".

    Args:
        seed (int32/int64): seed for random number generator

    Returns:
        initializer (Callable): callable that creates an initial value from a given shape
    """
    return tf.initializers.glorot_normal(seed)


def orthogonal_init(gain: float = 1.0, seed=None) -> Callable:
    """ Orthogonal initializer

    If the shape of the tensor to initialize is two-dimensional, it is initialized
    with an orthogonal matrix obtained from the QR decomposition of a matrix of
    random numbers drawn from a normal distribution.

    If the matrix has fewer rows than columns then the output will have orthogonal
    rows. Otherwise, the output will have orthogonal columns.

    If the shape of the tensor to initialize is more than two-dimensional,
    a matrix of shape `(shape[0] * ... * shape[n - 2], shape[n - 1])`
    is initialized, where `n` is the length of the shape vector.
    The matrix is subsequently reshaped to give a tensor of the desired shape.

    !!! cite "References"
        1. [Exact solutions to the nonlinear dynamics of learning in deep linear neural networks](https://openreview.net/forum?id=_wzZwKpTDF_9C)

    Args:
        gain (float): multiplicative factor to apply to the orthogonal matrix
        seed (int32/int64): seed for random number generator

    Returns:
        initializer (Callable): callable that creates an orthogonal matrix from a given shape
    """

    return tf.initializers.orthogonal(gain=gain, seed=seed)


def identity_init(gain: float = 1.0):
    """ Identity Initializer

    creates an identity matrix for a 2D shape

    Args:
        gain (float): multiplicative factor to be applied to the identity matrix

    Returns:
        initializer (Callable): callable that creates an identity matrix from a given 2D shape
    """
    return tf.initializers.identity(gain=gain)


def he_uniform_init(seed=None):
    """ He Uniform Initializer

    also known as `MSRA` initialization

    It draws samples from a uniform distribution within $[-l, l]$ where $l = \\sqrt{\\frac{6}{fan_{in}}}$ where
    $fan_{in}$ is the number of input units in the weight tensor.

    !!! Cite "References"
        1. [Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification](https://arxiv.org/abs/1502.01852)

    Args:
        seed (int32/int64): seed for random number generator

    Returns:
        initializer (Callable): callable that returns a tensor value from a given shape

    """
    return tf.initializers.he_uniform(seed=seed)


def he_normal_init(seed=None):
    """ He Normal Initializer

    also known as `MSRA` initialization

    It draws samples from a truncated normal distribution centered on $0$ with
    $stddev = \\sqrt{\\frac{2}{fan_{in}}} where $fan_{in}$ is the number of input units in the weight tensor.


    !!! Cite "References"
        1. [Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification](https://arxiv.org/abs/1502.01852)

    Args:
        seed (int32/int64): seed for random number generator

    Returns:
        initializer (Callable): callable that returns a tensor value from a given shape

    """
    return tf.initializers.he_normal(seed=seed)


def variance_scaling_init(scale=1.0, mode="fan_in", uniform=False, seed=None):
    """ variance scaling init

    A generalization of `glorot_uniform_init` and `he_uniform_init`.
    The principle behind this is that it's good to to keep the scale of the input variance constant,
    so it does not explode or diminish by reaching the final layer.

    ```
    if mode='fan_in': # Count only number of input connections.
        n = fan_in
    elif mode='fan_out': # Count only number of output connections.
        n = fan_out
    elif mode='fan_avg': # Average number of inputs and output connections.
        n = (fan_in + fan_out)/2.0

    truncated_normal(shape, 0.0, stddev=sqrt(scale / n))
    ```

    !!! note "Example Configurations"
        * `factor=2.0, mode='FAN_IN', uniform=False`: MSRA / He normal initialization
        * `factor=1.0 mode='FAN_AVG' uniform=True/False`: glorot / xavier initialization

    Args:
        mode (`str`): "fan_in", "fan_out" or "fan_avg".
        scale (`float`): positive scaling factor
        uniform (`bool`): if True samples from uniform distribution, else from the truncated normal
        seed (int32/int64): seed for random number generator

    Returns:
        initializer (Callable): callable that returns a tensor value from a given shape

    """
    return tf.initializers.VarianceScaling(scale=scale, mode=mode,
                                           distribution="uniform" if uniform else "truncated_normal", seed=seed)



import tensorflow as tf
from tensorx.utils import as_tensor
from typing import Union, List
from tensorflow import Tensor, SparseTensor


def tensor_equal(first, second):
    """ returns a `bool` tensor with value `True` if tensors are equal,
    `False` otherwise.

    Args:
        first: An object whose type has a registered `Tensor` conversion function.
        second: Another object whose type has a registered `Tensor` conversion function.

    Returns:
        tensor (`Tensor`): a tensor with dtype `tf.bool`

    """
    first: Union[Tensor, SparseTensor] = as_tensor(first)
    second: Union[Tensor, SparseTensor] = as_tensor(second)
    if first.dtype != second.dtype:
        return tf.constant(False)
    if type(first) != type(second):
        if isinstance(first, SparseTensor):
            first = tf.sparse.to_dense(first)
        elif isinstance(second, SparseTensor):
            second = tf.sparse.to_dense(second)
    elif isinstance(first, SparseTensor):
        diff: SparseTensor = tf.abs(tf.sparse.add(first, second * tf.cast(-1, second.dtype)))
        first = diff.values
        second = tf.zeros_like(diff.values)

    if len(first.shape.as_list()) != len(second.shape.as_list()):
        return tf.constant(False)

    # logic_and evaluates everything so we use cond instead
    return tf.cond(
        tf.reduce_all(tf.equal(tf.shape(first), tf.shape(second))),
        true_fn=lambda: tf.reduce_all(tf.equal(first, second)),
        false_fn=lambda: tf.constant(False)
    )


def same_shape(tensor1, tensor2):
    return tensor_equal(tf.shape(tensor1), tf.shape(tensor2))


def shape_equal(shape1, shape2):
    shape1 = tf.TensorShape(shape1)
    shape2 = tf.TensorShape(shape2)
    return all([d1 == d2 for d1, d2 in zip(shape1, shape2)])


def tensor_close(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):  # pylint: disable=missing-docstring
    rtol_ = as_tensor(rtol)
    atol_ = as_tensor(atol)
    result = (tf.math.abs(a - b) <= atol_ + rtol_ * tf.math.abs(b))
    if equal_nan:
        result = tf.logical_or(result, tf.logical_and(tf.math.is_nan(a), tf.math.is_nan(b)))
    return result


def tensor_all_close(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
    return tf.reduce_all(tensor_close(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan))



import tensorflow as tf
import tensorx as tx
from tensorx.utils import Graph
import logging
from tensorx.train.callbacks import *
import numpy as np
from typing import Union, Dict
from queue import Empty
import os
import csv
import re

logger = logging.getLogger('tensorx')


class Model:
    """ Base Model
    Args:
        TODO convert callable to layer
        train_loss: either a callable, layer, or dictionary mapping a callable or
        layer to the target outputs.

        train_inputs: defaults to run inputs, if loss is provided you can either
        supply the inputs to the train graph that include the loss, or let
        the Model create inputs for you.
    """

    def __init__(self,
                 run_outputs,
                 run_inputs=None,
                 train_outputs=None,
                 train_inputs=None,
                 train_loss=None,
                 eval_inputs=None,
                 eval_outputs=None,
                 eval_score=None,
                 name='Model',
                 ):
        self.name = name

        self.run_inputs = as_list(run_inputs)
        self.run_outputs = as_list(run_outputs)
        self.train_inputs = as_list(train_inputs)
        self.train_outputs = as_list(train_outputs)
        self.eval_inputs = as_list(eval_inputs)
        self.eval_outputs = as_list(eval_outputs)
        self.eval_score = as_list(eval_score)

        if not isinstance(train_loss, tx.Layer):
            raise TypeError(f"Invalid train_loss type\n"
                            f"\t expected: Layer"
                            f"\t actual: {type(train_loss)}")

        self.train_loss = train_loss

        # layer graph -> function
        self.compiled = dict()

        # TODO problem with missing inputs in recurrent neural networks, solved by allowing missing inputs
        #  and supplying the inputs when doing as_function
        self.run_graph: Graph = Graph.build(inputs=run_inputs,
                                            outputs=run_outputs,
                                            add_missing_inputs=True)
        self.train_graph: Graph = Graph.build(inputs=self.train_inputs,
                                              outputs=self.train_outputs + [self.train_loss],
                                              add_missing_inputs=True)
        self.eval_graph: Graph = Graph.build(inputs=self.eval_inputs,
                                             outputs=self.eval_outputs + self.eval_score,
                                             add_missing_inputs=True)

        self.graph_inputs = {self.run_graph: self.run_inputs,
                             self.train_graph: self.train_inputs,
                             self.eval_graph: self.eval_inputs}

        # TODO add the possibility of having multiple optimizers that can be switched
        self.optimizer = None

        # optimizer: Param:
        self.optimizer_params = dict()
        # model properties accessible to callbacks
        self.optimization_step = dict()

        self.model_props = set()

    def draw(self, path="graph.pdf"):
        # TODO add edges for shared state
        try:
            from pygraphviz import AGraph

            def add_graph(g: AGraph, layer_graph: tx.Graph, cluster):
                for node in layer_graph.nodes:
                    # HTML for record nodes https://graphviz.org/doc/info/shapes.html#top
                    g.add_node(f"{cluster}_{node.name}", shape="none", margin=0, label=tx.utils.vizstyle(node))
                for node in layer_graph.nodes:
                    for other_node in layer_graph.edges_out[node]:
                        g.add_edge(f"{cluster}_{node.name}", f"{cluster}_{other_node.name}")

            dg = AGraph(directed=True)
            if self.run_graph and self.run_graph.nodes:
                dg.add_subgraph(name="cluster_run", label="run")
                g = dg.subgraphs()[-1]
                add_graph(g, self.run_graph, cluster="run")
            if self.train_graph and self.train_graph.nodes:
                dg.add_subgraph(name="cluster_train", label="train")
                g = dg.subgraphs()[-1]
                add_graph(g, self.train_graph, "train")
            if self.eval_graph and self.eval_graph.nodes:
                dg.add_subgraph(name="cluster_eval", label="eval")
                g = dg.subgraphs()[-1]
                add_graph(g, self.eval_graph, cluster="eval")

            dg.layout(prog="dot")
            dg.draw(path=path)

        except ImportError:
            raise ImportError("Could't find required pygraphviz module")

    @property
    def trainable_variables(self):
        return list(itertools.chain(*(layer.trainable_variables for layer in self.train_graph.nodes)))

    #
    # TODO I wonder if I can get just one optimizer per model
    #   the use case would be train a model n steps with an optimizer then use a callback to switch
    #   to a different optimizer etc.
    #   In this case the train loop would still be a single loop but the optimizers would switch inside of it
    #   I would need the option to switch optimizers manually
    #   but I still have a single loss
    #   Use case 2.
    #   -how to switch between two losses ?
    #   -I guess different losses would use different optimization steps since we need to gather the gradients
    #    for them
    #   - in keras you set the loss with compile, but where it connects along with its inputs becomes fuzzy
    #   - unless we have a way to set the optimizer to a specific loss?
    #   - since the loss is a layer it can have a name but these names are not guaranteed to be unique
    #   (I can force this)
    #   - the goal is not to use string names but object so we can have optimizers connected to specific losses
    #   - if multiple losses are defined, setting an optimizer will create a join loss, otherwise a loss must be
    #   specified
    #   - the model subclass should be responsible for exposing names for the loss references if necessary
    #   SET OPTIMIZER
    #   1. if no loss specified, create join loss with MEAN layer
    #   2. if loss specified create optimize step for that loss
    #   3. train will call specific optimizer or all of them in sequence? for the same data? this is a problem
    #   I guess I can ignore the later feature for now
    def set_optimizer(self, optimizer, **config):
        """ Set the optimizer for this model

        !!! Note "Optimizer Hyper-Parameters"
            The arguments passed to the optimizer constructor can be either regular Python values,
            tensors, or a `callable`. If they are callable, they will called during apply_gradients()
            to get the value for the hyper parameter.

        Args:
            optimizer (Optimizer): optimizer class or instance
            **config : dictionary with parameters for the optimizer, if you want to modify these parameters during
            training pass an ``tx.Param`` as the value for the given parameter instead of constant value.

        Returns:
            optimizer (Optimizer): the configured optimizer instance.
        """
        if isinstance(optimizer, tf.optimizers.Optimizer):
            # attributes can be set on optimizers
            # https://github.com/tensorflow/tensorflow/blob/25006be096cd4f0242a3b979fb212bbd192127b3/tensorflow/python/keras/optimizer_v2/optimizer_v2.py#L553
            for attr in config:
                setattr(optimizer, attr, config[attr])
        elif issubclass(optimizer, tf.optimizers.Optimizer):
            optimizer = optimizer.from_config(config)
        else:
            raise TypeError(f"optimizer_cls should be an instance of subclass of tf.optimizers.Optimizer:\n"
                            f"\tgot {type(optimizer)} instead.")

        self.optimizer = optimizer
        self.optimizer_params[optimizer] = dict()

        # different optimizers might have the same param names with different values
        for param_name, param in config.items():
            if isinstance(param, tx.Param):
                self.optimizer_params[optimizer][param_name] = param

        if self.train_graph not in self.compiled:
            self.compiled[self.train_graph] = self.train_graph.as_function(ord_inputs=self.train_inputs)
        train_fn = self.compiled[self.train_graph]

        @tf.function
        def optimization_step(*data):
            with tf.GradientTape() as tape:
                *train_out, loss = train_fn(*data)
                cfg = self.optimizer.get_config()

                grads = tape.gradient(loss, self.trainable_variables)

                # TODO add debugging when gradient is None
                #  most likely this happens when a variable is not dependent on the given function
                # for g, var in zip(grads, self.trainable_variables):
                #    if g is None:
                #        print(var.name)

                if "clipnorm" in cfg:
                    clipnorm = cfg["clipnorm"]
                    grads = [tf.clip_by_norm(g, clipnorm) for g in grads]
                if "clipvalue" in cfg:
                    clipvalue = cfg["clipvalue"]
                    grads = [tf.clip_by_value(g, -clipvalue, clipvalue) for g in grads]

                self.optimizer.apply_gradients(zip(grads, self.trainable_variables))

                return train_out + [loss]

        self.optimization_step[optimizer] = optimization_step

        return optimizer

    def run(self, input_feed, compiled_graph=False):
        if input_feed is not None:
            # TODO can models have params to be changed by input feed?
            params = self.model_props
            data_feed, param_feed = Model.parse_input(input_feed, self.run_inputs, params)

            # feed all params if necessary
            if param_feed:
                for param_name in param_feed:
                    params[param_name].value = param_feed[param_name]

        if not compiled_graph:
            return self.run_graph(data_feed)
        else:
            if self.run_graph not in self.compiled:
                self.compiled[self.run_graph] = self.run_graph.as_function(ord_inputs=self.run_inputs)

        params = list(data_feed.values())
        return self.compiled[self.run_graph](*params)

    @staticmethod
    def parse_input(input_feed, ord_inputs, param_dict=None):
        # TODO this has to be model dependent? or receive the input nodes for the current model instead of graph
        #  because the only thing we take from graph is in_nodes
        """ parses input_feed into data_feed ordered by current graph in_node order and param_feed

        Args:
            input_feed: a dictionary from Layers or string to values, or a list with a value for each
                input in a given graph by the same order these are defined. ``Input`` layer keys map to graph inputs,
                ``str`` keys map to either optimizer parameters or model properties

            ord_inputs: the ordered inputs to be matched with input_feed dict or list order
            param_dict: dict of string to Properties/Params, with params we're matching with input_feed keys

        Returns:
            data_feed, param_feed, dictionaries the first from ``Input`` layers to values to be fed to these inputs, the
            second from ``str`` values to scalar values to be given to either the current Optimizer or model properties
        """
        if not isinstance(input_feed, dict):
            input_feed = as_list(input_feed)
            inputs = [x for x in ord_inputs if isinstance(x, tx.Input) and not x.constant]
            input_feed = dict(zip(inputs, input_feed))

        data_feed = dict()
        param_feed = dict()

        for in_node in input_feed:
            if in_node in ord_inputs:
                data_feed[in_node] = input_feed[in_node]
            elif isinstance(in_node, str):
                if (param_dict is not None and in_node in param_dict) or param_dict is None:
                    param_feed[in_node] = input_feed[in_node]
            else:
                if not isinstance(in_node, (str, tx.Input)):
                    raise TypeError(f"invalid type fed to model {type(in_node)}:\n"
                                    f"\texpected either Input Layer or str")
                else:
                    raise ValueError(f"{str(in_node)} not found neither in graph nor params")

        if len(data_feed) != len(ord_inputs):
            raise ValueError(f"model required {len(ord_inputs)} inputs, {len(data_feed)} found")

        # order data_feed
        data_feed = {in_node: data_feed[in_node] for in_node in ord_inputs}

        return data_feed, param_feed

    def train_step(self, input_feed):
        if input_feed is not None:
            params = self.optimizer_params[self.optimizer]
            data_feed, param_feed = Model.parse_input(input_feed, self.train_inputs, params)

            # feed all params if necessary
            if param_feed:
                for param_name in param_feed:
                    params[param_name].value = param_feed[param_name]

            if data_feed:
                for lr in data_feed:
                    if isinstance(lr, tx.Input) and not lr.constant:
                        lr.value = data_feed[lr]

        optimization_step = self.optimization_step[self.optimizer]
        feed_values = list(data_feed.values())
        return optimization_step(*feed_values)

    def eval_step(self, input_feed):
        """

        Args:
            input_feed:

        Returns:
            *eval_output, eval_score ((eval outputs,eval score)):
        """
        if input_feed is not None:
            data_feed, param_feed = Model.parse_input(input_feed, self.eval_inputs)

            eval_graph = self.eval_graph
            if eval_graph not in self.compiled:
                self.compiled[eval_graph] = eval_graph.as_function(ord_inputs=self.eval_inputs)

            static_eval_graph = self.compiled[eval_graph]
            # *eval_out, eval_score = eval_fn(*list(data_feed.values()))
            feed_values = list(data_feed.values())
            return static_eval_graph(*feed_values)
        else:
            return None

    def train(self, train_data, validation_data=None, test_data=None, epochs=1, steps_per_epoch=None, callbacks=[]):
        """ Main training loop

        Args:
            train_data: an iterable of dictionaries from Input Layers to values {Input:data}.
            (calling iter on this object should yield an iterator for an epoch.)

            validation_data: an iterable of dictionaries from Input Layers to values {Input:data}.
            test_data: an iterable of dictionaries from Input Layers to values {Input:data}.

            epochs (int): number of training epochs.
            steps_per_epoch: number of steps in an epoch, if not None, epochs are incremented each time
            this number of steps pass even if the entire train_data has not been transversed.

            callbacks: ``Callback`` functions scheduled during the training.

        """
        # train loop properties
        step = Property("step", 0)
        epoch = Property("epoch", 1)
        epoch_step = Property("epoch_step", 0)
        last_loss = Property("last_loss", 0)
        train_loss = Property("train_loss", None)
        total_epochs = StaticProperty("total_epochs", value=epochs)
        test_loss = Property("test_loss", None)

        # TODO training loop only works with a single optimizer
        #   changing the optimizer during this loop is problematic
        #   without replacing the properties in the scheduler for the new
        #   optimizer
        optimizer_props: dict = self.optimizer_params[self.optimizer]

        properties = [step,
                      epoch_step,
                      epoch,
                      train_loss,
                      last_loss,
                      test_loss,
                      total_epochs
                      ] + list(optimizer_props.values())

        scheduler = Scheduler(model=self,
                              properties=properties)

        for callback in callbacks:
            scheduler.register(callback)

        if steps_per_epoch is not None and train_data is not None:
            epoch_data = iter(train_data)

        if validation_data:
            validation_cb = Eval(target_property="validation_loss",
                                 dataset=validation_data,
                                 priority=-2)
            scheduler.register(validation_cb)

        if test_data:
            test_cb = Eval(property="test_loss",
                           dataset=test_data,
                           priority=-1)
            scheduler.register(test_cb)

        # MAIN TRAINING LOOP
        scheduler.trigger(OnLoop(AT.START))

        try:
            while epoch.value <= epochs:
                # EPOCH START
                # restart iterator for an epoch
                if steps_per_epoch is None and train_data is not None:
                    epoch_data = iter(train_data)

                epoch_step.value = 0
                total_loss = 0
                scheduler.trigger(OnEpoch(epoch.value, AT.START))
                while steps_per_epoch is None or epoch_step.value < steps_per_epoch:
                    try:
                        if train_data is not None:
                            feed_dict = next(epoch_data)
                        else:
                            feed_dict = {}

                        feed_dict, param_feed = Model.parse_input(feed_dict, self.train_inputs)

                        optimizer_props = self.optimizer_params[self.optimizer]
                        for param_name in param_feed:
                            if param_name in optimizer_props:
                                optimizer_props[param_name].value = param_feed[param_name]

                        # updated here because we want this property to give us the current step, to know when
                        # a step ends use OnEveryStep(at=AT.END)
                        epoch_step.value += 1
                        step.value += 1
                        # update property values
                        for param_name in param_feed:
                            # add new properties if not in the scheduler
                            if param_name not in scheduler.props:
                                scheduler.observe(Property(name=param_name, value=param_feed[param_name]))
                            else:
                                # only update property if value changes
                                prop = scheduler.props[param_name]
                                # only change the value if this value is different, no need to trigger
                                # redundant updates on the properties that don't change value
                                if prop.value != param_feed[param_name]:
                                    prop.value = param_feed[param_name]

                        scheduler.trigger(OnStep(step.value, AT.START))
                        scheduler.trigger(OnEpochStep(epoch_step.value, AT.START))

                        # TODO test if the loss is properly retrieved from the graph
                        *outputs, loss = self.train_step(feed_dict)
                        if not np.isscalar(loss):
                            if isinstance(loss, list):
                                loss = np.mean([np.mean(l) for l in loss])
                            else:
                                loss = np.mean(loss)
                        last_loss.value = loss
                        total_loss += loss
                        train_loss.value = total_loss / epoch_step.value

                        scheduler.trigger(OnStep(epoch_step.value, AT.END))
                        scheduler.trigger(OnEpochStep(epoch_step.value, AT.END))
                    except StopIteration:
                        break

                # EPOCH END
                scheduler.trigger(OnEpoch(epoch.value, AT.END))
                epoch.value += 1
        except StopTrain as e:
            logging.info("Training stopped: {}".format(str(e)))
        except Exception as e:
            logging.exception("Error: " + str(e))
            raise e

        scheduler.trigger(OnLoop(AT.END))


class StopTrain(Exception):
    pass


class Eval(Callback):
    def __init__(self, target_property="eval", eval_fn=None, dataset=None, trigger=OnEveryEpoch(at=AT.END), priority=1):
        """ Evaluation Callback

        Takes a data sample evaluation function and a dataset and returns the average evaluation value for
        the entire dataset.

        Args:
            target_property: name for the property created by this callback
            eval_fn: function applied to the average evaluation value before updating the property
            dataset: the dataset on which the model will be evaluated
            trigger: trigger for when the evaluation is run
            priority: callback priority
        """
        self.target_property = Property(name=target_property)
        self.dataset = dataset
        self.eval_fn = eval_fn

        def eval_fn(model, _):
            dataset_it = iter(self.dataset)
            sum_eval = 0
            steps = 0

            for feed_dict in dataset_it:
                # if feed dict is not a feed dict, it should be a tuple or another iterable
                if not isinstance(feed_dict, dict):
                    inputs = model.eval_graph.in_nodes
                    feed_dict = dict(zip(inputs, feed_dict))

                # *outputs, score
                outputs = model.eval_step(feed_dict)
                *_, mean_eval = outputs
                if not np.isscalar(mean_eval):
                    mean_eval = np.mean(mean_eval)
                sum_eval += mean_eval
                steps += 1

            avg_eval = sum_eval / steps
            if self.eval_fn:
                self.target_property.value = self.eval_fn(avg_eval)
            else:
                self.target_property.value = avg_eval

        super().__init__(trigger_dict={trigger: eval_fn},
                         properties=[self.target_property],
                         priority=priority)


class EarlyStop(Callback):
    """ EarlyStop Callback

    Executes each time a given target property changes

    if the callback depends on another callback that computes a given property, make sure EarlyStop
    is executed after this callback.

    Warning:
        if this relies on a property doesn't exist, it throws an error and interrupts the train loop.


    """

    def __init__(self,
                 patience=3,
                 lesser_better=True,
                 threshold=0.001,
                 target="validation_loss",
                 trigger=None, priority=1):
        self.patience = patience
        self.threshold = threshold,
        self.target = target
        self.last_eval = None
        self.patience_tick = 0

        def early_stop(model, properties):
            if self.target not in properties:
                raise KeyError(
                    "EarlyStop callback is trying to access a property that doesn't exist: {}".format(self.target))
            measure = properties[target].value
            if self.last_eval is None:
                self.last_eval = measure
            else:
                improvement = self.last_eval - measure
                if not lesser_better:
                    improvement = -1 * improvement

                if improvement < self.threshold:
                    self.patience_tick += 1
                    if self.patience_tick == self.patience:
                        raise StopTrain("the model didn't improve for the last {} epochs".format(patience))
                else:
                    self.patience_tick = 0

                self.last_eval = measure

        if trigger is None:
            trigger = OnValueChange(self.target)

        super().__init__(trigger_dict={trigger: early_stop}, priority=priority)


class StopOnNaN(Callback):
    """ StopOnNaN Callback

    Interrupts the training loop if the last_loss property returns NaN

    Note:
        the only event triggered after this is ``OnTrain(AT.END)``
    """

    def __init__(self):
        def raise_stop(model, properties):
            step = properties["epoch_step"].value
            epoch = properties["epoch"].value
            loss = properties["last_loss"].value
            if np.isnan(loss):
                raise StopTrain("loss returned NaN on epoch {epoch} step {step}".format(epoch=epoch, step=step))

        super().__init__({OnValueChange("last_loss"): raise_stop})


class Progress(Callback):
    """ Progress Callback

    creates a CLI progress bar for the training loop
    """

    def __init__(self, total_steps=None, monitor="train_loss", priority=1):
        self.total_steps = total_steps
        self.progress = None
        self.monitor = as_list(monitor)

        # optional, only tries to import it when we create a progress callback

        from tqdm.autonotebook import tqdm
        self.tqdm = tqdm

        def progress_init(model, properties):
            self.progress = self.tqdm(total=self.total_steps)

        def progress_step(model, properties):
            postfix = {}
            for prop_name in self.monitor:
                if prop_name not in properties:
                    raise KeyError(
                        "Progress callback is trying to access a property that doesn't exist: {}".format(self.monitor))
                prop = properties[prop_name].value
                postfix[prop_name] = prop
            self.progress.update(1)
            if len(postfix) > 0:
                self.progress.set_postfix(postfix)

        def progress_stop(model, properties):
            self.progress.close()

        trigger_dict = {OnLoop(AT.START): progress_init,
                        OnEveryStep(at=AT.END): progress_step,
                        OnLoop(AT.END): progress_stop}

        def progress_update(model, properties):
            if self.total_steps is None:
                self.total_steps = self.progress.n * properties["total_epochs"].value
                self.progress.total = self.total_steps

        if self.total_steps is None:
            trigger_dict[OnEveryEpoch(at=AT.END)] = progress_update

        super().__init__(trigger_dict=trigger_dict,
                         priority=priority)


class NewProperty(Callback):
    """ NewProperty Callback

    Creates a new property by applying a given function to the value of an existing property.

    """

    def __init__(self, target, fn, new_prop, init_val=None, triggers=OnEveryStep(), priority=1):
        self.property = Property(name=new_prop, value=init_val)
        self.fn = fn
        self.triggers = as_list(triggers)

        def update_prop(model, properties):
            if target not in properties:
                raise ValueError("Property callback tried to access property that doesn't exist: {}".format(target))
            prop = properties[target]
            self.property.value = self.fn(prop.value)

        trigger_dict = {trigger: update_prop for trigger in self.triggers}

        super().__init__(trigger_dict=trigger_dict, priority=priority, properties=self.property)


class LambdaCallback(Callback):
    """ Lambda Callback

    Executes a given function on a given trigger

    """

    def __init__(self, fn=None, triggers=OnEveryEpoch(at=AT.END), properties=[], priority=1):
        self.triggers = as_list(triggers)
        self.properties = properties
        self.fn = fn
        trigger_dict = {trigger: fn for trigger in triggers}

        super().__init__(trigger_dict=trigger_dict, priority=priority, properties=self.properties)


class DictLogger(Callback):
    """ DictLogger Callback

    Logs the values of the given properties to a dictionary accessible through the
    ``logs` attribute

    Args:
        props (List[str]): list of strings with names of properties to log
        trigger (callbacks.Event): even on which this callback is executed
        priority (int): callback priority (lower values have higher priority)

    Attributes:
        logs (dict): dictionary mapping property names (str) to lists of values
    """

    def __init__(self, props, trigger=OnEveryEpoch(at=AT.END), priority=1):
        self.props = as_list(props)
        self.logs = {name: [] for name in self.props}

        def get_props(model, properties):
            for prop_name in self.props:
                if prop_name not in properties:
                    raise KeyError("DictLogger tried to access a property that doesn't exist: {}".format(prop_name))
                prop = properties[prop_name]
                self.logs[prop_name].append(prop.value)

        super().__init__(trigger_dict={trigger: get_props}, priority=priority)


class DecayAfter(Callback):
    """ DecayAfter Callback

    Decays a given property using a given decay rate after a given number of epochs. This callback checks the model
    "epoch" property set in the main training loop, if epoch is greater or equal than ``decay_after`` then it decays
    a target property value using decay_rate.

    Args:
        decay_after (int): epoch after which the decay starts affecting a target property
        target_property (str): a name for the property to be changed
        decay_rate (float): the rate by which the property is changed with `value * decay_rate``
        decay_threshold (float): minimum value the target property will take after which the decay has no effect.
        priority (int) callback priority (lower values have higher priority)

    """

    def __init__(self,
                 decay_after=0,
                 target_property="learning_rate",
                 decay_rate=1.0,
                 decay_threshold=1e-6,
                 priority=1):
        self.target_property = target_property
        self.decay_rate = decay_rate
        self.decay_threshold = decay_threshold
        self.decay_after = decay_after

        def update_fn(_, properties):
            if self.target_property not in properties:
                raise KeyError(
                    "DecayAfter callback is trying to change a property that doesn't exist: {}".format(
                        self.target_property))
            epoch = properties["epoch"].value

            if epoch >= self.decay_after:
                prop = properties[self.target_property]
                prop.value = max(self.decay_threshold, prop.value * decay_rate)

        super().__init__(trigger_dict={OnEveryEpoch(at=AT.END): update_fn},
                         properties=[],
                         priority=priority)


class PlateauDecay(Callback):
    """ PlateauDecay Callback

    Decays a ``target`` property with a given ``decay_rate`` when a ``monitor`` property value plateaus: doesn't improve
    from the previously observed value.

    Note: patience param is reset after decay is applied

    Args:
        monitor (str): the measure on which we want to measure the improvement, by default "validation_loss"
        target (str): the property to be changed by this callback
        decay_threshold (float):value representing the difference between evaluations necessary for the update to occur
        decay_rate (float): rate through which the param value is reduced `(value = value * decay_rate)`
        decay_threshold: point beyond witch the param value is not reduced `max(value * decay_rate, decay_threshold)`
        less_is_better: if True, evaluation is considered to improve if it decreases, else it is considered to improve
        if it increases

    Attributes:
        eval_history (list): list with the evaluation values passed through the update function
        improvement_threshold (float): the necessary difference between evaluations for the update to occur
        decay_rate (float): rate through which the param value is reduced `(value = value * decay_rate)`
        decay_threshold (float): point beyond witch ``target`` is not reduced `max(value * decay_rate, decay_threshold)`
    """

    def __init__(self,
                 monitor,
                 target,
                 improvement_threshold=1.0,
                 less_is_better=True,
                 patience=1,
                 decay_rate=1.0,
                 decay_threshold=1e-6,
                 priority=1):
        self.improvement_threshold = improvement_threshold
        self.decay_rate = decay_rate
        self.decay_threshold = decay_threshold
        self.less_is_better = less_is_better
        self.monitor = monitor
        self.target = target
        self.eval_history = []
        self.patience_tick = 0
        self.patience = patience

        def update_fn(_, properties):
            # get properties
            evaluation = properties[self.monitor]
            to_change = properties[self.target]

            self.eval_history.append(evaluation.value)

            if len(self.eval_history) > 1:
                evaluation = 0
                if len(self.eval_history) > 1:
                    evaluation = self.eval_history[(-2 - self.patience_tick)] - self.eval_history[-1]
                    if not self.less_is_better:
                        evaluation = -1 * evaluation

                if evaluation < self.improvement_threshold:
                    self.patience_tick += 1
                    if self.patience_tick == self.patience:
                        to_change.value = max(to_change.value * self.decay_rate, self.decay_threshold)
                        self.patience_tick = 0
                else:
                    self.patience_tick = 0

        super().__init__(trigger_dict={OnEveryEpoch(at=AT.END): update_fn},
                         properties=[],
                         priority=priority)


class CSVLogger(Callback):
    """ CSVLogger Callback

    logs property values to a csv file.

    Args:
        monitors (List[str]): list of property names to be logged
        static_logs (dict): a dictionary of values (str to value) to be output along with the target properties
    """

    def __init__(self, monitors, out_filename, static_logs=None, trigger=OnEveryEpoch(at=AT.END), priority=1):
        self.property_names = as_list(monitors)
        self.static_logs = static_logs
        self.out_filename = out_filename
        self.out_file = None
        self.n = 0
        self.writer = None

        def get_props(props):
            props = {prop_name: props[prop_name] for prop_name in self.property_names}
            all_props = {p.name: p.value for p in props.values()}
            # add values from the provided logs
            all_props.update(self.static_logs)
            return all_props

        def log_init(model, properties):
            self.out_file = open(self.out_filename, "w")
            all_props = get_props(properties)
            self.writer = csv.DictWriter(f=self.out_file, fieldnames=all_props.keys())
            self.writer.writeheader()

        def log_clean(model, properties):
            self.out_file.close()

        def log(model, properties):
            all_props = get_props(properties)
            self.writer.writerow(all_props)
            self.out_file.flush()
            self.n += 1

        super().__init__(trigger_dict={OnStep(1, at=AT.START): log_init,
                                       trigger: log,
                                       OnLoop(at=AT.END): log_clean},
                         properties=None,
                         priority=priority)


class ResetState(LambdaCallback):
    """ ResetState Callback

    calls reset state on the model on the given triggers
    """

    def __init__(self,
                 triggers=OnEveryEpoch(at=AT.END),
                 priority=1):
        triggers = as_list(triggers)

        def reset(model: Model, _):
            model.reset_state()

        super().__init__(fn=reset, triggers=triggers, priority=priority)


class Plot(Callback):
    """ Callback that plots the given properties in real time

    """

    # noinspection PyBroadException
    def __init__(self, monitor,
                 cols=3,
                 backend='pyqtgraph',
                 keep_open=False,
                 save_plot=False,
                 output_file=None,
                 trigger=OnEveryEpoch(at=AT.END), priority=1):
        self.keep_open = keep_open
        self.monitor = set(as_list(monitor))
        self.output_file = output_file
        self.process = None
        self.queue = None
        self.stop_event = None
        self.backend = backend

        def plot_worker_mpl(queue, stop_event):
            out_file = self.output_file

            step = 1
            axs = {}

            num_props = len(self.monitor)
            rows = np.ceil(num_props / 3)

            if backend == "matplotlib":
                import matplotlib.pyplot as plt
                # only use the TkAgg backend if in console
                try:
                    from IPython import get_ipython
                    if 'IPKernelApp' not in get_ipython().config:  # pragma: no cover
                        raise ImportError("console")
                except Exception:
                    import matplotlib
                    matplotlib.use("TkAgg")

                plt.ion()
                fig = plt.figure()
                fig.canvas.toolbar.pack_forget()

                for i, prop_name in enumerate(self.monitor):
                    axs[prop_name] = fig.add_subplot(rows, cols, i + 1)  # i + 1)

            elif backend == "pyqtgraph":
                from pyqtgraph.Qt import QtGui
                import pyqtgraph as pg

                # pg.setConfigOption('background', 'w')
                app = QtGui.QApplication([])
                win = pg.GraphicsWindow(title="Properties")
                plots = {}
                text = {}
                # win.ci.setContentsMargins(20, 0, 0, 0)

                for i, prop_name in enumerate(self.monitor):
                    plot_item = win.addPlot(title=prop_name, row=i // cols, col=i % cols)
                    plots[prop_name] = plot_item
                    axs[prop_name] = plot_item.plot(pen=pg.mkPen('#FF2400', width=1))

                QtGui.QApplication.processEvents()
            else:
                raise ValueError("invalid backed, valid backend options: matplotlib, pyqtgraph")

            while True:
                try:
                    properties = queue.get_nowait()

                    for prop_name in properties.keys():
                        prop_value = properties[prop_name]
                        ax = axs[prop_name]

                        # if prop_value is None:
                        #    prop_value = 0

                        if self.backend == "matplotlib":
                            if step == 1:
                                xs = [step]
                                ys = [prop_value]
                            else:
                                line = ax.lines[0]
                                xs = np.append(line.get_xdata(), [step])
                                ys = np.append(line.get_ydata(), [prop_value])

                                line.set_xdata(xs)
                                line.set_ydata(ys)

                            ax.plot(xs[step - 1:], ys[step - 1:], color="#FF2400", linestyle="solid")

                            if step == 1:
                                plt.tight_layout()
                                ax.set_title(prop_name)
                        elif self.backend == "pyqtgraph":
                            xs, ys = ax.getData()
                            if prop_value is not None:
                                if xs is None:
                                    xs = []
                                    ys = []
                                elif len(xs) == 1:
                                    # create text label
                                    curve_point = pg.CurvePoint(ax)
                                    plots[prop_name].addItem(curve_point)
                                    label = pg.TextItem("test", color="#FF4200")
                                    label.setParentItem(curve_point)
                                    text[prop_name] = (curve_point, label)

                                xs = np.append(xs, [step])
                                ys = np.append(ys, [prop_value])

                                ax.setData(xs, ys)

                                if len(xs) > 1:
                                    # update text label
                                    curve_point, label = text[prop_name]
                                    curve_point: pg.CurvePoint = curve_point
                                    # curve point index relative to sample index (last in this case)
                                    curve_point.setIndex(step - 1)
                                    # anchor is relative to parent (top left corner)
                                    label.setAnchor((1.0, 1.0))
                                    label.setText('%0.3f' % (ys[-1]))

                    if self.backend == "matplotlib":
                        fig.canvas.draw_idle()
                        fig.canvas.flush_events()
                    else:
                        QtGui.QApplication.processEvents()

                    step += 1
                except Empty:
                    if stop_event.is_set():
                        break
                    else:
                        if self.backend == "pyqtgraph":
                            QtGui.QApplication.processEvents()

            if self.backend == "matplotlib":
                plt.ioff()
                if self.keep_open:
                    plt.show()
                if save_plot:
                    out_file = self.output_file if self.output_file is not None else "train_run_{}.pdf".format(
                        str(os.getpid()))

                    plt.savefig(out_file)
                plt.close(fig)
            else:

                if save_plot:
                    import pyqtgraph.exporters
                    exporter = pg.exporters.ImageExporter(win.scene())
                    if out_file is None:
                        out_file = "train_run_{}.pdf".format(str(os.getpid()))
                    exporter.export(out_file)
                if self.keep_open:
                    QtGui.QApplication.instance().exec_()
                app.closeAllWindows()

        def plot_init(model, properties):
            import multiprocessing as mp
            self.stop_event = mp.Event()
            self.queue = mp.Queue()
            self.process = mp.Process(target=plot_worker_mpl, args=(self.queue, self.stop_event,))
            self.process.start()

        def plot_props(model, properties):
            data = {}
            for prop_name in self.monitor:
                if prop_name not in properties:
                    raise KeyError("Plot callback tried to access a property that doesn't exist: {}".format(prop_name))

                y = properties[prop_name].value
                data[prop_name] = y
            self.queue.put(data)

        def plot_clean(model, properties):
            self.stop_event.set()
            self.process.join()

        super().__init__(trigger_dict={OnLoop(at=AT.START): plot_init,
                                       trigger: plot_props,
                                       OnLoop(at=AT.END): plot_clean}, priority=priority)



import tensorflow as tf
import tensorx as tx
from tensorx.layers import Lambda


def binary_cross_entropy(labels, logits, name="binary_cross_entropy"):
    """ Binary Cross Entropy

    Measures the probability error in discrete binary classification tasks in which each class is independent and
    not mutually exclusive.

    !!! note "On Entropy and Cross-Entropy"

        Entropy refers to the number of bits required to transmit a randomly selected event from a probability
        distribution. A skewed distribution has a low entropy, whereas a distribution where events have equal
        probability has a larger entropy.

        The entropy of a random variable with a set $x \\in X$ discrete states and their
        probability $P(x)$, can be computed as:

        $$
            H(X) = –\\sum_{x \\in X} P(x) * log(P(x))
        $$

        Cross-entropy builds upon this idea to compute the number of bits required to represent or
        transmit an average event from one distribution compared to another distribution. if we consider a target
        distribution $P$ and an approximation of the target distribution $Q$, the cross-entropy of $Q$ from $P$
        is the number of additional bits to represent an event using Q instead of P:

        $$
            H(P, Q) = –\\sum_{x \\in X} P(x) * log(Q(x))
        $$


    !!! warning
        This is to be used on the **logits** of a model, not on the predicted labels.
        See also [from TensorFlow](https://www.tensorflow.org/api_docs/python/tf/nn/sigmoid_cross_entropy_with_logits).

    Args:
        labels (`Tensor`): empiric probability values (labels that occurred for a given sample)
        logits (`Tensor`): unscaled log probabilities used to predict the labels with `sigmoid(logits)`
        name (str): op name


    Returns:
        tensor (`Tensor`): binary (sigmoid) cross-entropy loss.

    """
    return tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=logits, name=name)


def categorical_cross_entropy(labels, logits, axis=-1, name="categorical_cross_entropy"):
    """ Categorical Cross entropy

    Measures the probability error in discrete classification tasks in which the classes are mutually exclusive.

   !!! warning
        This is to be used on the **logits** of a model, not on the predicted labels. Do not call this loss with the
        output of softmax.
        See also [from TensorFlow](https://www.tensorflow.org/api_docs/python/tf/nn/softmax_cross_entropy_with_logits).

    Args:
        labels (Tensor): empiric probability distribution. Each row labels[i] must be a valid probability distribution
        (integrate to 1).
        logits (Tensor): unscaled log probabilities used to predict the labels with `softmax(logits)`
        axis (int): The class dimension. Defaulted to -1 which is the last dimension.
        name (str): op name

    Returns:
        tensor (`Tensor`): categorical (softmax) cross-entropy loss.

    """
    return tf.nn.softmax_cross_entropy_with_logits(labels=labels, logits=logits, axis=axis, name=name)


def binary_hinge(labels, logits):
    """ Binary Hinge Loss

    Measures the classification error for maximum-margin classification. Margin classifiers like
    Support Vector Machines (SVM) maximise the distance between the closest examples and the decision boundary
    separating the binary classes. The hinge loss is defined as:

    $$
    \\ell(y) = \\max(0, 1-t \\cdot y),
    $$

    where $t$ is the intended output (labels) and $y$ are the output logits from the
    classification decision function, not the predicted class label.

    Args:
        labels (`Tensor`): tensor with values -1 or 1. Binary (0 or 1) labels are converted to -1 or 1.
        logits (`Tensor`): unscaled log probabilities.

    Returns:
        tensor (`Tensor`): hinge loss float tensor
    """
    return tf.losses.hinge(labels, logits)


def mse(target, predicted):
    """ Mean Squared Error (MSE) Loss

    Measures the average of the squares of the errors - the difference between an estimator and what is estimated.
    This is a risk function, corresponding to the expected value of the quadratic loss:

    $$
    MSE =\\frac{1}{N}​\\sum^{N}_{i=0}​(y-\\hat{y})^2
    $$

    !!! info
        MSE is sensitive towards outliers and given several examples with the same input feature values,
        the optimal prediction will be their mean target value. This should be compared with _Mean Absolute
        Error_, where the optimal prediction is the median. MSE is thus good to use if you believe that your
        target data, conditioned on the input, is normally distributed around a mean value --and when it's
        important to penalize outliers.

    Args:
        predicted (`Tensor`): estimated target values
        target (`Tensor`): ground truth, correct values

    Returns:
        tensor (`Tensor`): mean squared error value

    """
    return tf.losses.mean_squared_error(target, predicted)


def sparsemax_loss(logits, labels, name="sparsemax_loss"):
    """ Sparsemax Loss

    A loss function for the sparsemax activation function. This is similar to `tf.nn.softmax`, but able to output s
    parse probabilities.

    !!! info
        Applicable to multi-label classification problems and attention-based neural networks
        (e.g. for natural language inference)

    !!! cite "References"
        1. [From Softmax to Sparsemax: A Sparse Model of Attention and Multi-Label Classification](https://arxiv.org/abs/1602.02068)


    Args:
        labels (`Tensor`): the target dense labels (one hot encoded)
        logits (`Tensor`): unnormalized log probabilities
        name (str): op name

    Returns:
        loss (`Tensor`): sparsemax loss
    """

    with tf.name_scope(name):
        logits = tf.convert_to_tensor(logits)
        sparsemax = tx.sparsemax(logits)
        labels = tf.convert_to_tensor(labels, name="labels")

        shifted_logits = logits - tf.math.reduce_mean(logits, axis=1)[:, tf.newaxis]

        # sum over support (support = predicted labels)
        support = tf.cast(sparsemax > 0, sparsemax.dtype)
        sum_s = support * sparsemax * (shifted_logits - 0.5 * sparsemax)

        # - z_k + ||q||^2
        q_part = labels * (0.5 * labels - shifted_logits)

        return tf.math.reduce_sum(sum_s + q_part, axis=1)


def kld(target, predicted):
    """ Kullback–Leibler Divergence Loss

    Kullback–Leibler divergence (also called relative entropy) is a measure of how one probability distribution is
    different from a second, reference probability distribution.

    $$
    D_{KL}(P || Q) = - \\sum_{x \\in X}P(x) log\\left(\\frac{Q(x)}{P(x)}\\right)
    $$

    it is the expectation of the logarithmic difference between the probabilities $P$ and $Q$, where the
    expectation is taken using the probabilities $P$.

    Args:
        target (`Tensor`): target probability distribution
        predicted (`Tensor`): distribution predicted by the model

    Returns:
        kld (`Tensor`): LK divergence between the target and predicted distributions
    """
    return tf.losses.kullback_leibler_divergence(target, predicted)


def sinkhorn_loss(target, predicted, epsilon, n_iter, cost_fn=None):
    """ Sinkhorn Loss

    Alias:
        * `tx.metrics.sinkhorn`

    !!! info
        Optimal Transport (OT) provides a framework from which one can define a more powerful geometry to compare
        probability distributions. This power comes, however, with a heavy computational price. The cost of computing OT
        distances scales at least in $O(d^3 log(d))$ when comparing two histograms of dimension $d$. Sinkhorn algorithm
        alleviate this problem by solving an regularized OT in linear time.

    Given two measures with n points each with locations x and y
    outputs an approximation of the Optimal Transport (OT) cost with regularization
    parameter epsilon, niter is the maximum number of steps in sinkhorn loop

    !!! cite "References"
        1. [Concerning nonnegative matrices and doubly stochastic matrices](https://msp.org/pjm/1967/21-2/p14.xhtml)
        2. [Sinkhorn Distances:Lightspeed Computation of Optimal Transport](https://papers.nips.cc/paper/4927-sinkhorn-distances-lightspeed-computation-of-optimal-transport.pdf)

    Args:
        predicted (`Tensor`): model distribution
        target (`Tensor`): ground_truth, empirical distribution
        epsilon (float): regularization term >0
        n_iter (int): number of sinkhorn iterations
        cost_fn (Callable): function that returns the cost matrix between y_pred and y_true, defaults to $|x_i-y_j|^p$.

    Returns:
        cost (`Tensor`): sinkhorn cost of moving from the mass from the model distribution `y_pred` to the empirical
        distribution `y_true`.
    """
    return tx.sinkhorn(target, predicted, epsilon=epsilon, n_iter=n_iter, cost_fn=cost_fn)


class Loss(Lambda):
    pass


class BinaryCrossEntropy(Loss):
    def __init__(self, labels, logits, name="BinaryCrossEntropy"):
        super().__init__(labels, logits, fn=binary_cross_entropy, name=name)


class CategoricalCrossEntropy(Loss):
    def __init__(self, labels, logits, name="CategoricalCrossEntropy"):
        super().__init__(labels, logits, fn=categorical_cross_entropy, name=name)


class Sinkhorn(Loss):
    # TODO default values for n_iter and epsilon
    def __init__(self, y_pred, y_true, n_iter=1, epsilon=1e-1, cost_fn=None, name="Sinkhorn"):
        super().__init__(y_pred, y_true,
                         fn=lambda pred, true: sinkhorn_loss(pred, true, epsilon=epsilon, n_iter=n_iter,
                                                             cost_fn=cost_fn),
                         name=name)


class MSE(Loss):
    def __init__(self, target, predicted, name="MSE"):
        super().__init__(target, predicted, fn=mse, name=name)


__all__ = [
    "binary_cross_entropy",
    "categorical_cross_entropy",
    "BinaryCrossEntropy",
    "CategoricalCrossEntropy",
    "mse",
    "MSE"
]



import tensorflow as tf
from typing import Optional, List
from tensorx.utils import as_tensor
from tensorx.math import logit


def gumbel_top(logits, num_samples, dtype=tf.int32, seed=None):
    """ gumbel_top sampling

    uses the Gumbel-Top trick to sample without replacement from a discrete probability distribution
    parameterized by given (possibly unnormalized) log-probabilities `logits`.

    Args:
        logits (`Tensor`): log probabilities parameterizing the discrete distribution
        num_samples (int): number of unique samples to draw from the distribution
        dtype (`DType`): output dtype
        seed (int): random seed

    Returns:
        samples (int): a tensor with the indices sampled from the target distribution with shape
            `[shape(logits)[0],num_samples]`.
    """

    with tf.name_scope("gumbel_top"):
        shape = tf.shape(logits)
        u = tf.random.uniform(shape, minval=0, maxval=1, seed=seed)
        g = -tf.math.log(-tf.math.log(u))
        z = u + g
        _, indices = tf.nn.top_k(z, k=num_samples)
        if indices.dtype != dtype:
            indices = tf.cast(indices, dtype)
        return indices


def bernoulli(shape, prob=0.5, seed=None, dtype=tf.float32, name="random_bernoulli"):
    """ Random bernoulli tensor.

    A bernoulli tensor sampled from a random uniform distribution.

    !!! info
        The Bernoulli distribution takes the value _1_ with probability p and the value 0 with probability _q = 1 − p_

    Args:
        shape (`List[int]`): output tensor shape
        prob (`float`): probability of each element being 1
        seed (`int`): A Python `int`. Used in combination with `tf.random.set_seed` to create a reproducible sequence of
            tensors across multiple calls.
        dtype (`DType`): output tensor type
        name (`Optional[str]`): name for random_bernoulli op

    Returns:
        tensor (`Tensor`): tensor with the given input shape
    """
    with tf.name_scope(name=name):
        # uniform [prob, 1.0 + prob)
        random_tensor = prob
        random_tensor += tf.random.uniform(
            shape, seed=seed, dtype=dtype)
        return tf.math.floor(random_tensor)


def sample_sigmoid(logits, n, dtype=None, seed=None, name="sample_sigmoid"):
    """ sample_sigmoid

    Efficient sampling Bernoulli random variable from a sigmoid defined distribution

    !!! info
        This can be applied to the output layer of a neural net if this represents a bernoulli
        distribution defined using a parameterized sigmoid-activated layer

    Args:
        logits (`Tensor`): logits
        n (`int`): number of samples per row of logits
        dtype (`tf.DType`): input Tensor dtype
        seed (`int`): random number generator seed
        name (`Optional[str]`): name for sample sigmoid op (optional)

    Returns:
        samples (`Tensor`): a tensor with samples
    """
    with tf.name_scope(name):
        logits = as_tensor(logits)
        if dtype is not None:
            logits = tf.cast(logits, dtype)
        n = as_tensor(n)

        shape = tf.shape(logits)
        sample_shape = tf.concat([[n], shape], axis=-1)

        uniform_sample = tf.random.uniform(sample_shape,
                                           minval=0,
                                           maxval=1,
                                           dtype=logits.dtype, seed=seed)
        z = logit(uniform_sample, dtype=logits.dtype)

        return tf.cast(tf.greater(logits, z), tf.float32)


__all__ = [
    "gumbel_top",
    "bernoulli",
    "sample_sigmoid"
]



from .__version__ import __version__

from tensorx.layers import *
from tensorx.random import *
from tensorx.ops import *
from tensorx.math import *
from tensorx.metrics import *
from tensorx.train import *
from tensorx.init import *
from tensorx.loss import *
from tensorx.activation import *
from tensorx.logic import *
